/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Command line argument processing. */
class Args {

    /** Command line option management. */
    protected static abstract class Option {
        /** JArgs command line parser. */
        protected final jargs.gnu.CmdLineParser m_jaParser;
        /** JArgs command line option. */
        protected final jargs.gnu.CmdLineParser.Option m_jaOption;
        private final String m_jShortDescription;
        private final String m_jDetailedDescription;

        private Option(
                jargs.gnu.CmdLineParser JA_Parser, jargs.gnu.CmdLineParser.Option JA_Option,
                String J_ShortDescription, String J_DetailedDescription) {
            m_jaParser = JA_Parser;
            m_jaOption = JA_Option;
            m_jShortDescription = J_ShortDescription;
            m_jDetailedDescription = J_DetailedDescription;
        }

        /** Option parsing handler. */
        protected abstract void parse();

        /** Short description accessor.
            @return The short option description. */
        public String getShortDescription() {
            return m_jShortDescription;
        }
        /** Detailed description accessor.
            @return The detailed option description. */
        public String getDetailedDescription() {
            return m_jDetailedDescription;
        }
        /** Boolean value accessor handler.
            @return Invalid value. */
        public boolean getbValue() {
            return false;
        }
        /** String value accessor handler.
            @return Invalid value. */
        public String getstrValue() {
            return null;
        }
    }
    /** Boolean command line option management. */
    protected static final class BooleanOption extends Option {
        private boolean m_bValue;

        private BooleanOption(
                jargs.gnu.CmdLineParser JA_Parser, jargs.gnu.CmdLineParser.Option JA_Option,
                String J_ShortDescription, String J_DetailedDescription) {
            super(JA_Parser, JA_Option, J_ShortDescription, J_DetailedDescription);
        }

        protected void parse() {
            m_bValue = (m_jaParser.getOptionValue(m_jaOption) != null);
        }
        public boolean getbValue() {
            return m_bValue;
        }
    }
    /** String command line option management. */
    protected static final class StringOption extends Option {
        private String m_jValue;

        private StringOption(
                jargs.gnu.CmdLineParser JA_Parser, jargs.gnu.CmdLineParser.Option JA_Option,
                String J_ShortDescription, String J_DetailedDescription) {
            super(JA_Parser, JA_Option, J_ShortDescription, J_DetailedDescription);
        }

        protected void parse() {
            m_jValue = (String) m_jaParser.getOptionValue(m_jaOption);
        }

        public String getstrValue() {
            return m_jValue;
        }
    }

    private final String m_jAppFullName;
    private final String m_jAppShortName;
    private final jargs.gnu.CmdLineParser m_jaParser;
    private final Option m_wHelp;
    private final java.util.Vector<Option> m_jOptions;
    private String[] m_arjInputFiles;
    private Option m_wRemainingArgs;

    private final Option m_wVerbose;
    private final Option m_wTraces;
    private final Option m_wQuiet;

    /** Constructor.
        @param J_AppShortName Application short name.
        @param J_AppFullName Application full name. */
    protected Args(String J_AppShortName, String J_AppFullName) {
        m_jAppFullName = J_AppFullName;
        m_jAppShortName = J_AppShortName;

        m_jaParser = new jargs.gnu.CmdLineParser();
        m_jOptions = new java.util.Vector<Option>();
        m_arjInputFiles = null;

        m_wHelp = new BooleanOption(m_jaParser, m_jaParser.addBooleanOption('h', "help"), "[{-h,--help}]", "-h,--help: Displays this help and quit");
        m_wVerbose = addBooleanOption('v', "verbose", "[{-v,--verbose}]", "-v,--verbose: Show verbose output information");
        m_wTraces = addStringOption('t', "traces", "[{-t,--traces} trace1[,trace2...]]", "-t,--traces trace1[,trace2...]: Traces activation (coma seperated: " + Traces.getClasses() + ")");
        m_wQuiet = addBooleanOption('q', "quiet", "[{-q,--quiet}]", "-q,--quiet: Disable warnings");
        m_wRemainingArgs = new StringOption(m_jaParser, null, "[input file ...]", "[input file ...]: List of input files");
    }
    /** Remaining arguments description.
        @param J_ShortDescription Short description for remaining arguments.
        @param J_DetailedDescription Detailed description for remaining arguments. */
    protected void setRemainingArgsDescription(String J_ShortDescription, String J_DetailedDescription) {
        m_wRemainingArgs = new StringOption(m_jaParser, null, J_ShortDescription, J_DetailedDescription);
    }
    /** Boolean option addition.
        @param C_Option Single character option name.
        @param J_Option Full option name.
        @param J_ShortDescription Short description of the option.
        @param J_DetailedDescription Detailed description of the option.
        @return The new boolean option. */
    protected Option addBooleanOption(char C_Option, String J_Option, String J_ShortDescription, String J_DetailedDescription) {
        Option w_Option = new BooleanOption(m_jaParser, m_jaParser.addBooleanOption(C_Option, J_Option), J_ShortDescription, J_DetailedDescription);
        m_jOptions.add(w_Option);
        return w_Option;
    }
    /** String option addition.
        @param C_Option Single character option name.
        @param J_Option Full option name.
        @param J_ShortDescription Short description of the option.
        @param J_DetailedDescription Detailed description of the option.
        @return The new string option. */
    protected Option addStringOption(char C_Option, String J_Option, String J_ShortDescription, String J_DetailedDescription) {
        Option w_Option = new StringOption(m_jaParser, m_jaParser.addStringOption(C_Option, J_Option), J_ShortDescription, J_DetailedDescription);
        m_jOptions.add(w_Option);
        return w_Option;
    }

    /** Full name of application accessor.
        @return The full name of the application. */
    public String getAppFullName() { return m_jAppFullName; }
    /** Short name of application handler.
        @return The short name of the application. */
    public String getAppShortName() { return m_jAppShortName; }

    /** Displays a help message explaining the usage of the program. */
    protected void printUsage() {
        java.util.Iterator<Option> j_It = null;

        Display.err().put(getAppFullName() + " version " + Watch.VERSION + " " + Watch.COPYRIGHT + " " + Watch.AUTHOR + " " + Watch.EMAIL).endl();
        Display.err().endl();

        Display.err().put("Usage: " + getAppShortName() + " " + m_wHelp.getShortDescription()).endl();
        Display.err().put("Usage: " + getAppShortName() + " ");
        for (j_It = m_jOptions.iterator(); j_It.hasNext(); ) {
            Option w_Option = j_It.next();
            if (w_Option != null) {
                Display.err().put(w_Option.getShortDescription());
            }
            Display.err().put(" ");
        }
        Display.err().put(m_wRemainingArgs.getShortDescription()).endl();

        Display.err().put("    " + m_wHelp.getDetailedDescription()).endl();
        for (j_It = m_jOptions.iterator(); j_It.hasNext(); ) {
            Option w_Option = j_It.next();
            if (w_Option != null) {
                Display.err().put("    " + w_Option.getDetailedDescription()).endl();
            }
        }
        Display.err().put("    " + m_wRemainingArgs.getDetailedDescription()).endl();
    }

    /** Argument parsing.
        @param ARJ_Args List of arguments as received by a main() function. */
    public void parse(String[] ARJ_Args) {
        try {
            m_jaParser.parse(ARJ_Args);
            m_wHelp.parse();
            for (java.util.Iterator<Option> j_It = m_jOptions.iterator(); j_It.hasNext(); ) {
                Option w_Option = j_It.next();
                if (w_Option != null) {
                    w_Option.parse();
                }
            }
            m_arjInputFiles = m_jaParser.getRemainingArgs();

            if (m_wHelp.getbValue()) {
                printUsage();
                System.exit(0);
            }

            Display.showWarnings(! getbQuiet());
            for (java.util.Iterator<String> j_TraceFilterIt = getTraces().iterator(); j_TraceFilterIt.hasNext(); ) {
                Traces.showTraces(j_TraceFilterIt.next(), true);
            }

            for (java.util.Iterator<String> j_InputFileNameIt = getInputFiles().iterator(); j_InputFileNameIt.hasNext(); ) {
                Traces.traceApp(" " + j_InputFileNameIt.next());
            }
            Traces.traceApp("input files count = " + m_arjInputFiles.length);
            Traces.traceApp("-v,--verbose = " + new Boolean(getbVerbose()));
            Traces.traceApp("-q,--quiet = " + new Boolean(getbQuiet()));
            for (java.util.Iterator<String> j_TraceFilterIt = getTraces().iterator(); j_TraceFilterIt.hasNext(); ) {
                Traces.traceApp("-t,--traces = " + j_TraceFilterIt.next());
            }

        } catch (jargs.gnu.CmdLineParser.OptionException e) {
            Display.printStackTrace(e);
            printUsage();
            System.exit(2);
        }
    }

    /** Verbose option value accessor.
        @return The verbose option. */
    public boolean getbVerbose() {
        Traces.traceApp("Args.getbVerbose() = " + new Boolean(m_wVerbose.getbValue()));
        return m_wVerbose.getbValue();
    }
    /** Quiet option value accessor.
        @return The quiet option. */
    public boolean getbQuiet() {
        Traces.traceApp("Args.getbQuiet() = " + new Boolean(m_wQuiet.getbValue()));
        return m_wQuiet.getbValue();
    }
    /** Traces option settings accessor.
        @return The Traces option settings. Each string set in the vector represent a selected trace class. */
    public java.util.Vector<String> getTraces() {
        java.util.Vector<String> j_Vector = new java.util.Vector<String>();
        String j_Traces = m_wTraces.getstrValue();
        if (j_Traces != null) {
            String[] arj_Traces = j_Traces.split(",");
            if (arj_Traces != null) {
                for (int i=0; i<arj_Traces.length; i++) {
                    j_Vector.add(arj_Traces[i]);
                }
            }
        }
        Traces.traceApp("Args.getTraces() = " + j_Vector);
        return j_Vector;
    }
    /** Input files accessor.
        @return Input files list. */
    public java.util.Vector<String> getInputFiles() {
        java.util.Vector<String> j_Vector = new java.util.Vector<String>();
        for (int i=0; i<m_arjInputFiles.length; i++) {
            j_Vector.add(m_arjInputFiles[i]);
        }
        Traces.traceApp("Args.getInputFiles() = " + j_Vector);
        return j_Vector;
    }
}

