/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Computation model class. */
class Computation {

    /** Constructor.
        @param W_Date                   Date of the computation.
        @param D_Reference              Reference value at this moment of the computation.
        @param D_ReferenceRisks         Reference risks value at this moment of the computation.
        @param D_ReferenceOpportunities Reference opportunities at this moment of the computation.
        @param D_Imputed                Imputed value at this moment of the computation.
        @param D_Remaining              Remaining value at this moment of the computation.
        @param D_RemainingRisks         Remaining risks value at this moment of the computation.
        @param D_RemainingOpportunities Remaining opportunities value at this moment of the computation.
        @param D_Planned                Planned imputations value at this moment of the computation.
        @param D_Earned                 Earned value (against reference) at this moment of the computation. */
    public Computation(
            Date W_Date,
            double D_Reference,
            double D_ReferenceRisks,
            double D_ReferenceOpportunities,
            double D_Imputed,
            double D_Remaining,
            double D_RemainingRisks,
            double D_RemainingOpportunities,
            double D_Planned,
            double D_Earned) {
        m_wDate = new Date(W_Date);
        m_dReference = D_Reference;
        m_dReferenceRisks = D_ReferenceRisks;
        m_dReferenceOpportunities = D_ReferenceOpportunities;
        m_dImputed = D_Imputed;
        m_dRemaining = D_Remaining;
        m_dRemainingRisks = D_RemainingRisks;
        m_dRemainingOpportunities = D_RemainingOpportunities;
        m_dPlanned = D_Planned;
        m_dEarned = D_Earned;
    }
    /** String representation of the computation.
        @return The string representation of the computation. */
    public String toString() {
        return (
            "Computation "
            + "date = " + m_wDate + ", "
            + "reference = " + new Double(m_dReference) + ", "
            + "reference-risks = " + new Double(m_dReferenceRisks) + ", "
            + "reference-opportunities = " + new Double(m_dReferenceOpportunities) + ", "
            + "imputed = " + new Double(m_dImputed) + ", "
            + "remaining = " + new Double(m_dRemaining) + ", "
            + "remaining-risks = " + new Double(m_dRemainingRisks) + ", "
            + "remaining-opportunities = " + new Double(m_dRemainingOpportunities) + ", "
            + "planned = " + new Double(m_dPlanned) + ", "
            + "earned = " + new Double(m_dEarned)
        );
    }

    /** Date accessor.
        @return The date of the computation. */
    public Date getDate() {
        return new Date(m_wDate);
    }

    /** Reference value accessor.
        @return The reference value of the computation. */
    public double getReference() {
        return m_dReference;
    }

    /** Reference risks value accessor.
        @return The reference risks value of the computation. */
    public double getReferenceRisks() {
        return m_dReferenceRisks;
    }

    /** Reference opportunities value accessor.
        @return The reference opportunities value of the computation. */
    public double getReferenceOpportunities() {
        return m_dReferenceOpportunities;
    }

    /** Imputed value accessor.
        @return The imputed value of the computation. */
    public double getImputed() {
        return m_dImputed;
    }

    /** Remaining value accessor.
        @return The remaining value of the computation. */
    public double getRemaining() {
        return m_dRemaining;
    }

    /** Remaining risks value accessor.
        @return The remaining risks value of the computation. */
    public double getRemainingRisks() {
        return m_dRemainingRisks;
    }

    /** Remaining opportunities value accessor.
        @return The remaining opportunities value of the computation. */
    public double getRemainingOpportunities() {
        return m_dRemainingOpportunities;
    }

    /** Planned value accessor.
        @return The planned value of the computation. */
    public double getPlanned() {
        return m_dPlanned;
    }
    /** Earned value accessor.
        @return The earned value of the computation. */
    public double getEarned() {
        return m_dEarned;
    }

    private final Date m_wDate;
    private final double m_dReference;
    private final double m_dReferenceRisks;
    private final double m_dReferenceOpportunities;
    private final double m_dImputed;
    private final double m_dRemaining;
    private final double m_dRemainingRisks;
    private final double m_dRemainingOpportunities;
    private final double m_dPlanned;
    private final double m_dEarned;

}
