/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Date model class, providing parsing and formatting.
    All dates are aligned to midnight time.
    The public visibility is set for unit tests purpose. */
public class Date extends java.util.Date {

    /** Current locale.
        Default is the system default locale. */
    protected static java.util.Locale m_jCurrentLocale = java.util.Locale.getDefault();

    /** Current locale setting.
        @param J_CurrentLocale New current locale. Must not be null.
        @return true when success, false otherwise. */
    public static boolean setCurrentLocale(java.util.Locale J_CurrentLocale) {
        if (J_CurrentLocale != null) {
            m_jCurrentLocale = J_CurrentLocale;
            return true;
        }
        return false;
    }

    /** Constructor.
        @param J_Date Date in its string format. */
    public Date(String J_Date) {
        try {
            // First of all, try to parse the string using the current locale.
            setTime(
                java.text.DateFormat.getDateInstance(java.text.DateFormat.SHORT, m_jCurrentLocale)
                .parse(J_Date)
                .getTime()
            );
        } catch (java.text.ParseException j_ParseException) {
            // Then let's try to parse it as a 8601 date string.
            boolean b_Res = false;

            String[] arj_Split = J_Date.split("-");
            if (arj_Split.length >= 3) {
                int i_Year = new Integer(arj_Split[0]);
                int i_Month = new Integer(arj_Split[1]);
                int i_Day = new Integer(arj_Split[2]);

                try {
                    setTime(new java.util.GregorianCalendar(i_Year, i_Month-1, i_Day).getTime().getTime());
                    b_Res = true;
                } catch (Exception e) {
                    //Display.printStackTrace(e);
                }
            }

            if (! b_Res) {
                // No way, just display the original parse exception.
                Display.printStackTrace(j_ParseException);
            }
        } catch (Exception e) {
            // Display the exception.
            Display.printStackTrace(e);
        }
    }
    /** String representation of the date.
        Depends on the current locale.
        @return The string representation of the date. */
    public String toString() {
        try {
            return java.text.DateFormat.getDateInstance(java.text.DateFormat.SHORT, m_jCurrentLocale).format(this);
        } catch (Exception e) {
            return super.toString();
        }
    }

    /** Copy constructor (from a Watch date object).
        @param W_Date Watch date to copy. */
    public Date(Date W_Date) {
        setTime(W_Date.getTime());
    }

    /** Copy constructor (from a Java date object).
        @param J_Date Java date to copy. */
    public Date(java.util.Date J_Date) {
        setTime(J_Date.getTime());
    }

    /** Zero date accessor.
        The zero date corresponds to the very first date possible (usually Jan 1st 1970).
        @return Zero date. */
    public static Date zero() {
        Date w_Zero = today();
        w_Zero.setTime(0);
        return w_Zero;
    }
    /** Today date accessor.
        @return Date of the day. */
    public static Date today() {
        java.util.Calendar w_TodayCalendar = new java.util.GregorianCalendar();
        w_TodayCalendar.clear();
        w_TodayCalendar.set(
            java.util.Calendar.getInstance().get(java.util.Calendar.YEAR),
            java.util.Calendar.getInstance().get(java.util.Calendar.MONTH),
            java.util.Calendar.getInstance().get(java.util.Calendar.DAY_OF_MONTH)
        );
        return new Date(w_TodayCalendar.getTime());
    }
    /** Date incrementation.
        @return The date of the day after. */
    public Date nextDay() {
        long l_milliInDay = 24 * 60 * 60 * 1000;
        // Because some days don't increment if just the elapse time above is added (26/10/2008 for instance), let's add 4 hours more.
        // The date will be adjusted finally.
        l_milliInDay += 4 * 60 * 60 * 1000;
        Date w_NextDay = new Date(this);
        w_NextDay.setTime(this.getTime() + l_milliInDay);
        // Create a new date from the string representation in order to avoid derivations.
        return new Date(w_NextDay.toString());
    }
    /** Date decrementation.
        @return The date of the day before. */
    public Date prevDay() {
        long l_milliInDay = 24 * 60 * 60 * 1000;
        // See the comment in nextDay().
        l_milliInDay -= 4 * 60 * 60 * 1000;
        Date w_PrevDay = new Date(this);
        w_PrevDay.setTime(this.getTime() - l_milliInDay);
        // Create a new date from the string representation in order to avoid derivations.
        return new Date(w_PrevDay.toString());
    }

    /** Delay in days between two dates.
        @param W_Date1 First date.
        @param W_Date2 Second date.
        @return Number of days between the two dates. This number can be negative. */
    public static int delay(Date W_Date1, Date W_Date2) {
        if (W_Date2.before(W_Date1)) {
            return - delay(W_Date2, W_Date1);
        } else {
            int i_Days = 0;
            for (Date w_Date = W_Date1; w_Date.before(W_Date2); w_Date = w_Date.nextDay()) {
                i_Days ++;
            }
            return i_Days;
        }
    }

}
