/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Configuration file class. */
class EditorConfigFile {

    /** Default configuration file path.
        @return Default configuration file path. */
    public static String getDefaultPath() {
        return new java.io.File(System.getProperty("user.home"), ".WatchEditorConf.xml").getAbsolutePath();
    }


    private boolean m_bChanged;
    private String m_jLastDirectory;
    private String m_jCurrentInputFile;
    private final java.util.SortedSet<String> m_jProjects;
    // Window position.
    private boolean m_bWindowIsMaximized;
    private int m_iWindowX;
    private int m_iWindowY;
    private int m_iWindowWidth;
    private int m_iWindowHeight;

    /** Constructor. */
    public EditorConfigFile() {
        m_bChanged = false;
        m_jLastDirectory = null;
        m_jCurrentInputFile = null;
        m_jProjects = new java.util.TreeSet<String>();
        m_bWindowIsMaximized = false;
        m_iWindowX = 10;
        m_iWindowY = 10;
        m_iWindowWidth = EditorMainFrame.DEFAULT_WINDOW_WIDTH;
        m_iWindowHeight = EditorMainFrame.DEFAULT_WINDOW_HEIGHT;
    }

    /** Change notification. */
    public void setChanged() {
        m_bChanged = true;
    }

    /** Changed status accessor.
        @return Changed status. */
    public boolean getbChanged() {
        return m_bChanged;
    }

    /** Configuration file reset. */
    public void reset() {
        if (! m_jProjects.isEmpty()) {
            m_jProjects.clear();
            setChanged();
        }
    }

    /** Configuration file loading.
        @param J_FilePath Configuration file to load. */
    public void load(String J_FilePath) {
        reset();
        if (J_FilePath != null) {
            try {
                java.io.File j_File = new java.io.File(J_FilePath);
                if (! j_File.exists()) {
                    // Automatically create the config file.
                    new EditorXmlConfigFile().model2xml(j_File, this);
                } else {
                    // Read from the config file.
                    new EditorXmlConfigFile().xml2model(j_File, this);
                }
                m_bChanged = false;
            } catch (Exception e) {
                Display.printStackTrace(e);
            }
        }
    }

    /** Configuration file saving.
        @param J_FilePath Configuration file to save. */
    public void save(String J_FilePath) {
        if (J_FilePath != null) {
            try {
                new EditorXmlConfigFile().model2xml(new java.io.File(J_FilePath), this);
                m_bChanged = false;
            } catch (Exception e) {
                Display.printStackTrace(e);
            }
        }
    }

    /** Last directory accessor.
        @return The last directory path. */
    public String getLastDirectory() {
        if (m_jLastDirectory != null) { return m_jLastDirectory; }
        return new javax.swing.JFileChooser().getCurrentDirectory().getAbsolutePath();
    }

    /** Last directory setting.
        @param J_LastDirectory Last directory. */
    public void setLastDirectory(String J_LastDirectory) {
        if (    ((m_jLastDirectory != null) && (J_LastDirectory != null) && (! m_jLastDirectory.equals(J_LastDirectory)))
                || ((J_LastDirectory != null) && (m_jLastDirectory == null))
                || ((J_LastDirectory == null) && (m_jLastDirectory != null))) {
            m_jLastDirectory = J_LastDirectory;
            setChanged();
        }
    }

    /** Current input file accessor.
        @return The current input file. null if no current input file. */
    public String getCurrentInputFile() {
        if (m_jCurrentInputFile != null) { return m_jCurrentInputFile; }
        return null;
    }
    /** Safe current input file accessor.
        @return The current input file. Empty string if no current input file. */
    public String getSafeCurrentInputFile() {
        if (m_jCurrentInputFile != null) { return m_jCurrentInputFile; }
        return new String();
    }

    /** Current input file setting.
        @param J_CurrentInputFile Current input file. */
    public void setCurrentInputFile(String J_CurrentInputFile) {
        if (    ((m_jCurrentInputFile != null) && (J_CurrentInputFile != null) && (! m_jCurrentInputFile.equals(J_CurrentInputFile)))
                || ((J_CurrentInputFile != null) && (m_jCurrentInputFile == null) && (J_CurrentInputFile.length() > 0))
                || ((J_CurrentInputFile == null) && (m_jCurrentInputFile != null))) {
            m_jCurrentInputFile = J_CurrentInputFile;
            setChanged();
        }
    }

    /** Preferences projects accessor.
        @return List of preferences projects. */
    public java.util.SortedSet<String> getProjects() {
        return m_jProjects;
    }

    /** Reference project addition.
        @param J_ProjectPath New reference project path. */
    public void addProject(String J_ProjectPath) {
        getProjects().add(J_ProjectPath);
        setChanged();
        Editor.getInstance().reloadProjects();
    }

    /** Window position saving.
        @param B_IsMaximized true if the window is maximized.
        @param I_X X position of the window.
        @param I_Y Y position of the window.
        @param I_Width Width of the window.
        @param I_Height Heigt of the window. */
    public void saveWindowPosition(boolean B_IsMaximized, int I_X, int I_Y, int I_Width, int I_Height) {
        // Create a new instance for the default path.
        EditorConfigFile w_ConfigFile = new EditorConfigFile();
        w_ConfigFile.load(EditorConfigFile.getDefaultPath());

        // Save current window position.
        w_ConfigFile.setWindowIsMaximized(B_IsMaximized);
        w_ConfigFile.setWindowX(I_X);
        w_ConfigFile.setWindowY(I_Y);
        w_ConfigFile.setWindowWidth(I_Width);
        w_ConfigFile.setWindowHeight(I_Height);
        w_ConfigFile.save(EditorConfigFile.getDefaultPath());

        // And remember the window position in the main instance also.
        m_bWindowIsMaximized = B_IsMaximized;
        m_iWindowX = I_X;
        m_iWindowY = I_Y;
        m_iWindowWidth = I_Width;
        m_iWindowHeight = I_Height;
        // Do not consider the config file has changed.
    }
    /** Window maximized status accessor.
        @return The window maximized status. */
    public boolean getWindowIsMaximized() { return m_bWindowIsMaximized; }
    /** Window maximized status setter.
        @param B_IsMaximized New maximized status. */
    public void setWindowIsMaximized(boolean B_IsMaximized) { m_bWindowIsMaximized = B_IsMaximized; }
    /** X window position accessor.
        @return X window position. */
    public int getWindowX() { return m_iWindowX; }
    /** X window position setter.
        @param I_X New X window position. */
    public void setWindowX(int I_X) { m_iWindowX = I_X; }
    /** Y window position accessor.
        @return Y window position. */
    public int getWindowY() { return m_iWindowY; }
    /** Y window position setter.
        @param I_Y New Y window position. */
    public void setWindowY(int I_Y) { m_iWindowY = I_Y; }
    /** Window width accessor.
        @return Window width. */
    public int getWindowWidth() { return m_iWindowWidth; }
    /** Window width setter.
        @param I_Width New window width. */
    public void setWindowWidth(int I_Width) { m_iWindowWidth = I_Width; }
    /** Window height accessor.
        @return Window height. */
    public int getWindowHeight() { return m_iWindowHeight; }
    /** Window height setter.
        @param I_Height New window height. */
    public void setWindowHeight(int I_Height) { m_iWindowHeight = I_Height; }
}

