/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Editor input class. */
class EditorInput extends Watch {

    private Date m_wCurrentRevisionDate;
    private EditorImputationPanel m_wCurrentImputation;

    /** Constructor. */
    public EditorInput() {
        m_wCurrentRevisionDate = null;
        m_wCurrentImputation = null;
    }

    /** Comment input.
        @param J_Words Collection of words, making a sentence. */
    public void comment(java.util.Collection<String> J_Words) {
        String j_Comment = "";
        for (java.util.Iterator<String> j_It = J_Words.iterator(); j_It.hasNext(); ) {
            if (j_Comment.length() > 0) { j_Comment += " "; }
            String j_Word = j_It.next();
            if (j_Word != null) { j_Comment += j_Word; }
        }
        if (m_wCurrentImputation != null) {
            m_wCurrentImputation.addComment(j_Comment);
        } else {
            Display.err().put("EditorInput.comment(): No current imputation. Forgetting comment '" + j_Comment + "'").endl();
        }
    }

    /** Project name declaration.
        @param J_ProjectName New project name. */
    public void projectName(String J_ProjectName) {
        // Nothing to do
    }

    /** Locale declaration.
        @param J_Locale New locale in its string form.
        @see watch.Watch#locale */
    public void locale(String J_Locale) {
        // Do not prevent locale().
        super.locale(J_Locale);
    }

    /** Beginning of a task group.
        @param J_TaskId Task identifier.
        @param J_StartDate Task start date. */
    public void beginTaskGroup(String J_TaskId, String J_StartDate) {
        // Nothing to do, only final tasks are taken in account
    }

    /** End of a task group. */
    public void endTaskGroup() {
        // Nothing to do, only final tasks are taken in account
    }

    /** Single task definition.
        @param J_TaskId                 Task identifier.
        @param J_ReferenceStartDate     Reference start date. Can be null if no specific reference start date.
        @param D_Reference              Reference value for the task. -1.0 if this task should not be checked against its reference.
        @param D_ReferenceRisks         Reference risks of the task.
        @param D_ReferenceOpportunities Reference opportunities of the task.
        @param J_PlannedStartDate       Planned start date. Can be null if no planification is done.
        @param J_PlannedEndDate         Planned end date. Can be null if no planification is done.
        @param E_SchedulingMode         Scheduling mode. Either Task.SCHEDULING_NONE, Task.SCHEDULING_AT_END or Task.SCHEDULING_LINEAR. */
    public void singleTask(
            String J_TaskId,
            String J_ReferenceStartDate, double D_Reference, double D_ReferenceRisks, double D_ReferenceOpportunities,
            String J_PlannedStartDate, String J_PlannedEndDate, int E_SchedulingMode) {
        Date w_ReferenceStartDate = null; if (J_ReferenceStartDate != null) { w_ReferenceStartDate = new Date(J_ReferenceStartDate); }
        Date w_PlannedStartDate = null; if (J_PlannedStartDate != null) { w_PlannedStartDate = new Date(J_PlannedStartDate); }
        Date w_PlannedEndDate = null; if (J_PlannedEndDate != null) { w_PlannedEndDate = new Date(J_PlannedEndDate); }
        Editor.getInstance().addTask(new Task(
            J_TaskId,
            w_ReferenceStartDate, D_Reference, D_ReferenceRisks, D_ReferenceOpportunities,
            w_PlannedStartDate, w_PlannedEndDate, E_SchedulingMode
        ));
    }

    /** Current user setting. Useful for imputations for instance.
        @param J_UserName User identifier. */
    public void setCurrentUser(String J_UserName) {
        Editor.getInstance().getMainFrame().setUserName(J_UserName);
    }

    /** Imputation.
        @param J_TaskId Identifier of the task the imputation is related to.
        @param J_Date Date of the imputation.
        @param D_Imputed Imputation value. */
    public void imputation(String J_TaskId, String J_Date, double D_Imputed) {
        Date w_Date = new Date(J_Date);
        EditorDayPanel w_Day = Editor.getInstance().getMainFrame().getDay(w_Date);
        if (w_Day != null) {
            m_wCurrentImputation = w_Day.getImputation(J_TaskId);
            if (m_wCurrentImputation != null) {
                m_wCurrentImputation.setImputed(m_wCurrentImputation.getImputed() + D_Imputed);
            }
        }
    }

    /** Current revision date setting.
        @param J_Date Revision date to remember for next revisions and lock operations. */
    public void setCurrentRevisionDate(String J_Date) {
        m_wCurrentRevisionDate = new Date(J_Date);
    }

    /** Revision.
        @param J_TaskId Identifier of the task wich remaining value is being revised.
        @param J_Date Date of the revision.
        @param D_Remaining Revision of the remaining value.
        @param D_Risks Revision of the risks value.
        @param D_Opportunities Revision of the opportunities value.
        @param D_Earned Revision of the earned value. */
    public void revision(String J_TaskId, String J_Date, double D_Remaining, double D_Risks, double D_Opportunities, double D_Earned) {
        Date w_Date = m_wCurrentRevisionDate;
        if (J_Date != null) { w_Date = new Date(J_Date); }
        EditorDayPanel w_Day = Editor.getInstance().getMainFrame().getDay(w_Date);
        if (w_Day != null) {
            m_wCurrentImputation = w_Day.getImputation(J_TaskId);
            if (m_wCurrentImputation != null) {
                if (D_Remaining >= 0.0) { m_wCurrentImputation.setRemaining(D_Remaining); }
                if (D_Risks >= 0.0) { m_wCurrentImputation.setRisks(D_Risks); }
                if (D_Opportunities >= 0.0) { m_wCurrentImputation.setOpportunities(D_Opportunities); }
                if (D_Earned >= 0.0) { m_wCurrentImputation.setEarned(D_Earned); }
            }
        }
    }

    /** Lock operation.
        @param J_TaskId Identifier of the task being locked.
        @param J_Date Date of the lock operation. */
    public void lock(String J_TaskId, String J_Date) {
        Date w_Date = m_wCurrentRevisionDate;
        if (J_Date != null) { w_Date = new Date(J_Date); }
        EditorDayPanel w_Day = Editor.getInstance().getMainFrame().getDay(w_Date);
        if (w_Day != null) {
            m_wCurrentImputation = w_Day.getImputation(J_TaskId);
            if (m_wCurrentImputation != null) {
                m_wCurrentImputation.setLocked();
            }
        }
    }

    /** No executions. Disabling of Watch.execute().
        @param J_FocusTask Task to focus the computation onto.
        @param J_Users List of users to focus the computation onto.
        @param J_ReferenceDate Date to consider as the reference date.
        @param J_StartDate Date to start the computation from.
        @param J_EndDate Date to end the computation at.
        @param B_SumImputationsOnly When true, limits the computation at summing imputations.
        @param B_History When true, launches the history computation instead of a regular computation.
        @param W_Transformation Output transformation.
        @param J_OutputFileName Results output file name. */
    public void execute(
            String J_FocusTask, java.util.Collection<String> J_Users,
            String J_ReferenceDate, String J_StartDate, String J_EndDate,
            boolean B_SumImputationsOnly, boolean B_History,
            Transformation W_Transformation,
            String J_OutputFileName) {
        // Do nothing.
    }

    private void onError(String J_ErrorMessage) {
        javax.swing.JOptionPane.showMessageDialog(
            Editor.getInstance().getMainFrame(),
            J_ErrorMessage,
            "Error",
            javax.swing.JOptionPane.ERROR_MESSAGE
        );
    }

}
