/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Editor main frame.
    Contains editor application constants.
    The content of the frame constitutes the content of the document. Therefore, a changed status is directly managed by this class. */
class EditorMainFrame extends javax.swing.JFrame {

    /** Default window width. */
    public static final int DEFAULT_WINDOW_WIDTH = 800;
    /** Default window height. */
    public static final int DEFAULT_WINDOW_HEIGHT = 600;
    /** Default number field width (in characters). */
    public static final int NUMBER_WIDTH = 6;
    /** Default user name field width (in characters). */
    public static final int USER_WIDTH = 10;
    /** Default lock field width (in characters). */
    public static final int LOCK_WIDTH = 5;
    /** Default day field width (in characters). */
    public static final int DAY_WIDTH = 6;
    /** Default scroll field width (in characters).
        Hack for alignment of the imputation header panel with the scroll bar. */
    public static final int SCROLL_WIDTH = 1;

    /** Color of the background of the frame. */
    public static final java.awt.Color COLOR_BACKGROUND = java.awt.Color.WHITE;
    /** Color of the header day panels. */
    public static final java.awt.Color COLOR_DAY_TITLE = new java.awt.Color(0.95f, 0.95f, 1.0f);
    /** Color of the background of the imputation header panel. */
    public static final java.awt.Color COLOR_HEADER_BACKGROUND = new java.awt.Color(0.95f, 0.95f, 1.0f);
    /** Color of the foreground of the imputation header panel. */
    public static final java.awt.Color COLOR_HEADER_FOREGROUND = new java.awt.Color(0.75f, 0.75f, 1.0f);

    /** Static text field configuration.
        Size, alignment...
        @param SW_TextField Text field to configure.
        @param I_FieldWidth Required text field in characters. */
    public static void configureTextField(javax.swing.JTextField SW_TextField, int I_FieldWidth) {
        SW_TextField.setColumns(I_FieldWidth);
        SW_TextField.setMinimumSize(new java.awt.Dimension(SW_TextField.getPreferredSize().width, SW_TextField.getPreferredSize().height));
        SW_TextField.setMaximumSize(new java.awt.Dimension(SW_TextField.getPreferredSize().width, SW_TextField.getPreferredSize().height));
        SW_TextField.setHorizontalAlignment(javax.swing.JTextField.CENTER);
    }


    /** Constructor. */
    public EditorMainFrame() {
        super("Watch Editor");

        setWindowProperties();
        setMenuBar();
        setWindowContent();

        m_bChanged = false;
    }

    /** Changed status notification.
        Automatically updates the title of the frame. */
    public void setChanged() {
        if (! m_bChanged) {
            m_bChanged = true;
            updateWindowTitle();
        }
    }
    /** Unchanged status notifications.
        Automatically updates the title of the frame. */
    public void setUnchanged() {
        if (m_bChanged) {
            m_bChanged = false;
            updateWindowTitle();
        }
    }
    /** Changed status accessor.
        @return The changed status. */
    public boolean getbChanged() {
        return m_bChanged;
    }

    /** Updates the title of the window. */
    private void updateWindowTitle() {
        String j_Title = "Watch Editor";
        if (Editor.getInstance().getConfigFile().getCurrentInputFile() != null) {
            j_Title += (" - " + Editor.getInstance().getConfigFile().getCurrentInputFile());
        }
        if (Editor.getInstance().getMainFrame().getbChanged()) {
            j_Title += " *";
        }
        setTitle(j_Title);
    }

    /** Load complete notification.
        Useful at startup. Editor loads all input files given in the command line arguments, then calls this notification method. */
    public void loadComplete() {
        // In case of only one input file, remember the name of it
        if (Editor.getInstance().getArgs().getInputFiles().size() == 1) {
            Editor.getInstance().getConfigFile().setCurrentInputFile(Editor.getInstance().getArgs().getInputFiles().iterator().next());
        }
        setUnchanged();
        updateWindowTitle();

        // Display the window
        restoreWindowPosition();
        setVisible(true);
        if (getImputationsPanel() != null) {
            int i_DayCount = getImputationsPanel().getComponentCount();
            if (i_DayCount > 0) {
                java.awt.Component awt_Day = getImputationsPanel().getComponent(i_DayCount - 1);
                if (awt_Day instanceof javax.swing.JComponent) {
                    javax.swing.JComponent sw_Day = (javax.swing.JComponent) awt_Day;
                    sw_Day.scrollRectToVisible(sw_Day.getBounds());
                }
            }
        }
    }

    /** User name setting.
        Since the application manages only one user per frame, a warning message is displayed when several users are specified.
        @param J_UserName New user name. */
    public void setUserName(String J_UserName) {
        // If the username changes...
        if (    (! m_bOnlyOneUserWarningDisplayed)
                && (getUserPanel().getUserName() != null) && (getUserPanel().getUserName().length() > 0)
                && (J_UserName != null) && (J_UserName.length() > 0)
                && (! J_UserName.equals(getUserPanel().getUserName()))) {
            // ... and there are imputations
            if (getDays().size() > 0) {
                // ... display a warning message.
                javax.swing.JOptionPane.showMessageDialog(
                    Editor.getInstance().getMainFrame(),
                    "Several users specified by the input data "
                        + "(" + getUserPanel().getUserName() + ", " + J_UserName + "...) "
                        + "whereas the application manages only one!\r\n"
                        + "Please check before saving.",
                    "Warning",
                    javax.swing.JOptionPane.WARNING_MESSAGE
                );
                m_bOnlyOneUserWarningDisplayed = true;
            }
        }

        if ((J_UserName != null) && (J_UserName.length() > 0)) {
            getUserPanel().setUserName(J_UserName);
        }
    }
    /** Flag in order to display once the warning message indicating that only one user is managed by the application. */
    static private boolean m_bOnlyOneUserWarningDisplayed = false;

    /** User name accessor.
        @return The current user name. null if not set. */
    public String getUserName() {
        return getUserPanel().getUserName();
    }
    /** Safe user name accessor.
        @return The current user name. An empty string if not set. */
    public String getSafeUserName() {
        return getUserPanel().getSafeUserName();
    }

    /** Days of imputations accessor.
        @return The list of day panels, containing imputation panels each. */
    public java.util.Collection<EditorDayPanel> getDays() {
        java.util.Collection<EditorDayPanel> j_Days = new java.util.Vector<EditorDayPanel>();
        java.awt.Container awt_ImputationsPanel = getImputationsPanel();
        if (awt_ImputationsPanel != null) {
            for (int i=0; i<awt_ImputationsPanel.getComponentCount(); i++) {
                if (awt_ImputationsPanel.getComponent(i) instanceof EditorDayPanel) {
                    j_Days.add((EditorDayPanel) awt_ImputationsPanel.getComponent(i));
                }
            }
        }
        return j_Days;
    }

    /** Day panel accessor or creation on a given date.
        Searches for the given date in current day panels, and returns it when the date is found.
        Creates a new day panel otherwise.
        @param W_Date Searched date.
        @return The day panel researched or the new day panel. */
    public EditorDayPanel getDay(Date W_Date) {
        EditorDayPanel w_Day = null;
        for (java.util.Iterator<EditorDayPanel> j_It = getDays().iterator(); j_It.hasNext(); ) {
            w_Day = j_It.next();
            if (w_Day != null) {
                if (false) {}
                else if ((W_Date == null) && (w_Day.getDate() == null)) { return w_Day; }
                else if ((W_Date == null) || (w_Day.getDate() == null)) { continue; }
                else if (w_Day.getDate().equals(W_Date)) { return w_Day; }
            }
        }

        w_Day = new EditorDayPanel(W_Date);
        addDayPanel(w_Day);

        return w_Day;
    }

    /** Window property settings. */
    private void setWindowProperties() {
        // Window configuration
        setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
        addWindowListener(new java.awt.event.WindowAdapter() { public void windowClosing(java.awt.event.WindowEvent awt_Event) {
            Editor.getInstance().getMainFrame().onClose();
        }});
    }

    /** Window menu bar configuration.
        Creates menu, menu items and configure shortcuts. */
    private void setMenuBar() {
        javax.swing.JMenuBar sw_MenuBar = new javax.swing.JMenuBar(); setJMenuBar(sw_MenuBar);
            javax.swing.JMenu sw_FileMenu = sw_MenuBar.add(new javax.swing.JMenu("File"));
            sw_FileMenu.setMnemonic(java.awt.event.KeyEvent.VK_F);
                javax.swing.JMenuItem sw_FileOpenMenu = sw_FileMenu.add(new javax.swing.JMenuItem("Open"));
                    sw_FileOpenMenu.setMnemonic(java.awt.event.KeyEvent.VK_O);
                    sw_FileOpenMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_O, java.awt.event.InputEvent.CTRL_MASK));
                    sw_FileOpenMenu.setToolTipText("Open file");
                    sw_FileOpenMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        Editor.getInstance().getMainFrame().onOpen();
                    }});
                javax.swing.JMenuItem sw_EditClearMenu = sw_FileMenu.add(new javax.swing.JMenuItem("New"));
                    sw_EditClearMenu.setMnemonic(java.awt.event.KeyEvent.VK_N);
                    sw_EditClearMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_N, java.awt.event.InputEvent.CTRL_MASK));
                    sw_EditClearMenu.setToolTipText("New imputation file");
                    sw_EditClearMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        Editor.getInstance().getMainFrame().onClear();
                    }});
                javax.swing.JMenuItem sw_FileSaveMenu = sw_FileMenu.add(new javax.swing.JMenuItem("Save"));
                    sw_FileSaveMenu.setMnemonic(java.awt.event.KeyEvent.VK_S);
                    sw_FileSaveMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S, java.awt.event.InputEvent.CTRL_MASK));
                    sw_FileSaveMenu.setToolTipText("Save file");
                    sw_FileSaveMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        Editor.getInstance().getMainFrame().onSave();
                    }});
                javax.swing.JMenuItem sw_FileSaveAsMenu = sw_FileMenu.add(new javax.swing.JMenuItem("Save as"));
                    sw_FileSaveAsMenu.setMnemonic(java.awt.event.KeyEvent.VK_V);
                    sw_FileSaveAsMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S, java.awt.event.InputEvent.CTRL_MASK | java.awt.event.InputEvent.SHIFT_MASK));
                    sw_FileSaveAsMenu.setToolTipText("Save file as");
                    sw_FileSaveAsMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        Editor.getInstance().getMainFrame().onSaveAs();
                    }});
                javax.swing.JMenuItem sw_FileCloseMenu = sw_FileMenu.add(new javax.swing.JMenuItem("Exit"));
                    sw_FileCloseMenu.setMnemonic(java.awt.event.KeyEvent.VK_X);
                    sw_FileCloseMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_Q, java.awt.event.InputEvent.CTRL_MASK));
                    sw_FileCloseMenu.setToolTipText("Exit application");
                    sw_FileCloseMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        Editor.getInstance().getMainFrame().onClose();
                    }});
            javax.swing.JMenu sw_ConfigureMenu = sw_MenuBar.add(new javax.swing.JMenu("Preferences"));
            sw_ConfigureMenu.setMnemonic(java.awt.event.KeyEvent.VK_P);
                javax.swing.JMenuItem sw_ConfigureProjects = sw_ConfigureMenu.add(new javax.swing.JMenuItem("Add project"));
                    sw_ConfigureProjects.setMnemonic(java.awt.event.KeyEvent.VK_P);
                    sw_ConfigureProjects.setToolTipText("Add a project with its given task list");
                    sw_ConfigureProjects.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        Editor.getInstance().getMainFrame().onAddProject();
                    }});
            javax.swing.JMenu sw_EditMenu = sw_MenuBar.add(new javax.swing.JMenu("Edit"));
            sw_EditMenu.setMnemonic(java.awt.event.KeyEvent.VK_E);
                javax.swing.JMenuItem sw_EditTodayMenu = sw_EditMenu.add(new javax.swing.JMenuItem("Today"));
                    sw_EditTodayMenu.setMnemonic(java.awt.event.KeyEvent.VK_T);
                    sw_EditTodayMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_T, java.awt.event.InputEvent.CTRL_MASK));
                    sw_EditTodayMenu.setToolTipText("Edit today's imputations / revisions");
                    sw_EditTodayMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        EditorDayPanel w_Day = getDay(Date.today());
                        if (w_Day != null) {
                            w_Day.getImputation("");
                        }
                    }});
                javax.swing.JMenuItem sw_EditYesterdayMenu = sw_EditMenu.add(new javax.swing.JMenuItem("Yesterday"));
                    sw_EditYesterdayMenu.setMnemonic(java.awt.event.KeyEvent.VK_Y);
                    sw_EditYesterdayMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_Y, java.awt.event.InputEvent.CTRL_MASK));
                    sw_EditYesterdayMenu.setToolTipText("Edit yesterday's imputations / revisions");
                    sw_EditYesterdayMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        EditorDayPanel w_Day = getDay(Date.today().prevDay());
                        if (w_Day != null) {
                            w_Day.getImputation("");
                            w_Day.scrollRectToVisible(w_Day.getBounds());
                        }
                    }});
                javax.swing.JMenuItem sw_EditNewDayMenu = sw_EditMenu.add(new javax.swing.JMenuItem("New day"));
                    sw_EditNewDayMenu.setMnemonic(java.awt.event.KeyEvent.VK_D);
                    sw_EditNewDayMenu.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_D, java.awt.event.InputEvent.CTRL_MASK));
                    sw_EditNewDayMenu.setToolTipText("Edit imputations / revisions for a new date");
                    sw_EditNewDayMenu.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        EditorDayPanel w_Day = getDay(null);
                        if (w_Day != null) {
                            w_Day.getImputation("");
                            w_Day.scrollRectToVisible(w_Day.getBounds());
                        }
                    }});
            javax.swing.JMenu sw_HelpMenu = sw_MenuBar.add(new javax.swing.JMenu("Help"));
            sw_HelpMenu.setMnemonic(java.awt.event.KeyEvent.VK_H);
                javax.swing.JMenuItem sw_HelpAbout = sw_HelpMenu.add(new javax.swing.JMenuItem("About Watch Editor"));
                    sw_HelpAbout.setMnemonic(java.awt.event.KeyEvent.VK_A);
                    sw_HelpAbout.setToolTipText("About Watch Editor application");
                    sw_HelpAbout.addActionListener(new java.awt.event.ActionListener() { public void actionPerformed(java.awt.event.ActionEvent awt_Event) {
                        Editor.getInstance().getMainFrame().onAbout();
                    }});
    }

    /** Initial window content settings.
        Creates header panels, scroll bars... */
    private void setWindowContent() {
        // Window content
        getContentPane().setLayout(new javax.swing.BoxLayout(getContentPane(), javax.swing.BoxLayout.Y_AXIS));

        // Add the user panel.
        add(new EditorUserPanel());

        // Add the imputation header panel.
        add(new EditorImputationPanel.HeaderPanel());

        // Then add a JScrollPane.
        javax.swing.JScrollPane sw_ScrollPane = new javax.swing.JScrollPane(
            new javax.swing.JPanel(),
            javax.swing.JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
            javax.swing.JScrollPane.HORIZONTAL_SCROLLBAR_NEVER
        );
        sw_ScrollPane.getVerticalScrollBar().setUnitIncrement(16);
        add(sw_ScrollPane);

        // Set the imputations panel layout.
        java.awt.Container awt_ImputationsPanel = getImputationsPanel();
        awt_ImputationsPanel.setBackground(EditorMainFrame.COLOR_BACKGROUND);
        if (awt_ImputationsPanel != null) {
            awt_ImputationsPanel.setLayout(new javax.swing.BoxLayout(awt_ImputationsPanel, javax.swing.BoxLayout.Y_AXIS));
        }
    }

    /** Window position restoration. */
    private void restoreWindowPosition() {
        setBounds(
            Editor.getInstance().getConfigFile().getWindowX(), Editor.getInstance().getConfigFile().getWindowY(),
            Editor.getInstance().getConfigFile().getWindowWidth(), Editor.getInstance().getConfigFile().getWindowHeight()
        );
    }
    /** Window position saving. */
    private void saveWindowPosition() {
        Editor.getInstance().getConfigFile().saveWindowPosition(false, getX(), getY(), getWidth(), getHeight());
    }

    /** User panel accessor.
        @return The user panel. (null when the editor frame is not initialized yet). */
    private EditorUserPanel getUserPanel() {
        java.awt.Container awt_Container = getContentPane();
        if (awt_Container != null) {
            for (int i=0; i<awt_Container.getComponentCount(); i++) {
                if (awt_Container.getComponent(i) instanceof EditorUserPanel) {
                    return (EditorUserPanel) awt_Container.getComponent(i);
                }
            }
        }
        // default
        return null;
    }

    /** Imputation panel accessor.
        @return The imputations panel. (null when the editor frame is not initialized yet). */
    private java.awt.Container getImputationsPanel() {
        java.awt.Container awt_Container = getContentPane();
        if (awt_Container != null) {
            for (int i=0; i<awt_Container.getComponentCount(); i++) {
                if (awt_Container.getComponent(i) instanceof javax.swing.JScrollPane) {
                    java.awt.Component awt_ViewportView = ((javax.swing.JScrollPane) awt_Container.getComponent(i)).getViewport().getView();
                    if (awt_ViewportView instanceof java.awt.Container) {
                        return (java.awt.Container) awt_ViewportView;
                    }
                }
            }
        }
        // default
        return null;
    }

    /** Internal method for day panel addition.
        @param W_Day New day panel. */
    private void addDayPanel(EditorDayPanel W_Day) {
        java.awt.Container awt_ImputationsPanel = getImputationsPanel();
        if (awt_ImputationsPanel != null) {
            awt_ImputationsPanel.add(W_Day);
            validate();
            W_Day.scrollRectToVisible(W_Day.getBounds());

            Editor.getInstance().getMainFrame().setChanged();
        }
    }

    /** Handler called when the application is required to be closed.
        @return javax.swing.JOptionPane.YES_NO_CANCEL_OPTION. */
    private int onClose() {
        // Ask to save the imputation file.
        if (Editor.getInstance().getMainFrame().getbChanged()) {
            String j_Question = "Save?";
            if (Editor.getInstance().getConfigFile().getCurrentInputFile() != null) {
                j_Question = "Save " + Editor.getInstance().getConfigFile().getCurrentInputFile() +"?";
            }
            switch (javax.swing.JOptionPane.showConfirmDialog(
                Editor.getInstance().getMainFrame(),
                j_Question,
                "Save file",
                javax.swing.JOptionPane.YES_NO_CANCEL_OPTION)) {
            case javax.swing.JOptionPane.YES_OPTION:
                if (onSave() == javax.swing.JOptionPane.CANCEL_OPTION) {
                    return javax.swing.JOptionPane.CANCEL_OPTION;
                }
                break;
            case javax.swing.JOptionPane.CANCEL_OPTION:
                return javax.swing.JOptionPane.CANCEL_OPTION;
            }
        }

        // Ask to save the configuration file if it has changed.
        if (Editor.getInstance().getConfigFile().getbChanged()) {
            switch (javax.swing.JOptionPane.showConfirmDialog(
                Editor.getInstance().getMainFrame(),
                "Do you want to save the preferences?",
                "Save preferences",
                javax.swing.JOptionPane.YES_NO_CANCEL_OPTION)) {
            case javax.swing.JOptionPane.YES_OPTION:
                Editor.getInstance().getConfigFile().save(EditorConfigFile.getDefaultPath());
                break;
            case javax.swing.JOptionPane.CANCEL_OPTION:
                return javax.swing.JOptionPane.CANCEL_OPTION;
            }
        }

        // Eventually save the window position.
        saveWindowPosition();

        System.exit(0);
        return javax.swing.JOptionPane.YES_OPTION;
    }

    /** Handler called on the 'File/Open' menu.
        @return javax.swing.JOptionPane.YES_NO_CANCEL_OPTION. */
    private int onOpen() {
        javax.swing.JFileChooser sw_FileChooser = new javax.swing.JFileChooser(Editor.getInstance().getConfigFile().getLastDirectory());
        if (sw_FileChooser.showOpenDialog(Editor.getInstance().getMainFrame()) == javax.swing.JFileChooser.APPROVE_OPTION) {
            // Clear current data.
            onClear();

            // Then load the selected file.
            Editor.getInstance().loadInputFile(sw_FileChooser.getSelectedFile().getAbsolutePath());
            Editor.getInstance().getConfigFile().setLastDirectory(sw_FileChooser.getSelectedFile().getParentFile().getAbsolutePath());
            Editor.getInstance().getConfigFile().setCurrentInputFile(sw_FileChooser.getSelectedFile().getAbsolutePath());
            Editor.getInstance().getMainFrame().setUnchanged();

            return javax.swing.JOptionPane.YES_OPTION;
        } else {
            return javax.swing.JOptionPane.CANCEL_OPTION;
        }
    }

    /** Handler called on the 'File/Save' menu.
        @return javax.swing.JOptionPane.YES_NO_CANCEL_OPTION. */
    private int onSave() {
        // Overwrite current file ?
        if (Editor.getInstance().getConfigFile().getCurrentInputFile() != null) {
            Editor.getInstance().save(Editor.getInstance().getConfigFile().getCurrentInputFile());
            Editor.getInstance().getMainFrame().setUnchanged();
            return javax.swing.JOptionPane.YES_OPTION;
        } else {
            return onSaveAs();
        }
    }
    /** Handler called on the 'File/Save As' menu.
        @return javax.swing.JOptionPane.YES_NO_CANCEL_OPTION. */
    private int onSaveAs() {
        // Save as ?
        javax.swing.JFileChooser sw_FileChooser = new javax.swing.JFileChooser(Editor.getInstance().getConfigFile().getLastDirectory());
        if (sw_FileChooser.showOpenDialog(Editor.getInstance().getMainFrame()) == javax.swing.JFileChooser.APPROVE_OPTION) {
            Editor.getInstance().save(sw_FileChooser.getSelectedFile().getAbsolutePath());
            Editor.getInstance().getConfigFile().setLastDirectory(sw_FileChooser.getSelectedFile().getParentFile().getAbsolutePath());
            Editor.getInstance().getConfigFile().setCurrentInputFile(sw_FileChooser.getSelectedFile().getAbsolutePath());
            Editor.getInstance().getMainFrame().setUnchanged();
            return javax.swing.JOptionPane.YES_OPTION;
        } else {
            return javax.swing.JOptionPane.CANCEL_OPTION;
        }
    }

    /** Called on the 'Edit/Clear' menu or when the application needs to reset current data.
        @return javax.swing.JOptionPane.YES_NO_CANCEL_OPTION. */
    private int onClear() {
        // First of all, propose to save current data
        if (    (Editor.getInstance().getConfigFile().getCurrentInputFile() != null)
                && (Editor.getInstance().getMainFrame().getbChanged())) {
            switch (javax.swing.JOptionPane.showConfirmDialog(
                Editor.getInstance().getMainFrame(),
                "Save " + Editor.getInstance().getConfigFile().getCurrentInputFile() +"?",
                "Save file",
                javax.swing.JOptionPane.YES_NO_CANCEL_OPTION)) {
            case javax.swing.JOptionPane.YES_OPTION:
                if (onSave() == javax.swing.JOptionPane.CANCEL_OPTION) {
                    return javax.swing.JOptionPane.CANCEL_OPTION;
                }
                break;
            case javax.swing.JOptionPane.CANCEL_OPTION:
                return javax.swing.JOptionPane.CANCEL_OPTION;
            }
        }

        // Remove all components
        java.awt.Container awt_ImputationsPanel = getImputationsPanel();
        if (awt_ImputationsPanel != null) {
            for (int i = awt_ImputationsPanel.getComponentCount() - 1; i >= 0; i--) {
                if (awt_ImputationsPanel.getComponent(i) instanceof EditorDayPanel) {
                    awt_ImputationsPanel.remove(i);
                }
            }
        }
        // Reset the current file name
        Editor.getInstance().getConfigFile().setCurrentInputFile(null);
        Editor.getInstance().getMainFrame().setUnchanged();
        // Repaint the main frame.
        Editor.getInstance().getMainFrame().repaint();

        return javax.swing.JOptionPane.YES_OPTION;
    }

    /** Called on the 'Preferences/Add Project' menu.
        @return javax.swing.JOptionPane.YES_NO_CANCEL_OPTION. */
    private int onAddProject() {
        javax.swing.JFileChooser sw_FileChooser = new javax.swing.JFileChooser(Editor.getInstance().getConfigFile().getLastDirectory());
        if (sw_FileChooser.showOpenDialog(Editor.getInstance().getMainFrame()) == javax.swing.JFileChooser.APPROVE_OPTION) {
            Editor.getInstance().getConfigFile().addProject(sw_FileChooser.getSelectedFile().getAbsolutePath());
            Editor.getInstance().getConfigFile().setLastDirectory(sw_FileChooser.getSelectedFile().getParentFile().getAbsolutePath());
            return javax.swing.JOptionPane.YES_OPTION;
        } else {
            return javax.swing.JOptionPane.CANCEL_OPTION;
        }
    }

    /** Called on the 'Help/About' menu.
        @return javax.swing.JOptionPane.YES_NO_CANCEL_OPTION. */
    private int onAbout() {
        javax.swing.JOptionPane.showMessageDialog(
            this,
            Editor.getInstance().getArgs().getAppFullName() + " version " + Watch.VERSION + "\r\n"
                + Watch.COPYRIGHT + " " + Watch.AUTHOR + " " + Watch.EMAIL + "\r\n"
                + Watch.URL,
            "About " + Editor.getInstance().getArgs().getAppFullName(),
            javax.swing.JOptionPane.INFORMATION_MESSAGE
        );
        return javax.swing.JOptionPane.YES_OPTION;
    }


    /** Flag indicating whether the data have changed since opening or last saved. */
    private boolean m_bChanged;

}

