/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Imputation model class. */
class Imputation extends Commentable {

    /** Constructor.
        @param W_Task Task imputed.
        @param W_User User who has completed the imputation. If null, the imputation is not attached to any user.
        @param W_Date Date of the imputation.
        @param D_Imputed Value of the imputation. */
    public Imputation(Task W_Task, User W_User, Date W_Date, double D_Imputed) {
        m_wTask = W_Task;
        m_wUser = W_User;
        m_wDate = new Date(W_Date);
        m_dImputed = D_Imputed;
    }

    /** String representation of the imputation.
        @return The string representation of the imputation. */
    public String toString() {
        return (
            "Imputation "
            + "task = " + m_wTask + ", "
            + "user = " + m_wUser + ", "
            + "date = " + m_wDate + ", "
            + "imputed = " + new Double(m_dImputed)
        );
    }

    /** Imputed task accessor.
        @return The task imputed. */
    public Task getTask() {
        return m_wTask;
    }

    /** User who has completed the imputation accessor.
        @return The user who has completed the imputation. */
    public User getUser() {
        return m_wUser;
    }

    /** Date of imputation accessor.
        @return The date of the imputation. */
    public Date getDate() {
        return new Date(m_wDate);
    }

    /** Imputation value accessor.
        @return The value of the imputation. */
    public double getImputed() {
        return m_dImputed;
    }

    /** Imputation matching predicate.
        Determines whether the imputation corresponds to the user, the start and end date given.
        @param W_Users      User predicate. If null or empty, the predicate is not taken in account.
        @param W_StartDate  The date the imputation will be checked to be after. Must not be null.
        @param W_EndDate    The date the imputation will be checked to be before. Must not be null.
        @return true if all predicates are matched, false otherwise. */
    public boolean matches(User.Map W_Users, Date W_StartDate, Date W_EndDate) {
        if ((m_wUser == null) || ((W_Users != null) && (! W_Users.isEmpty()) && (! W_Users.containsKey(m_wUser.getName())))) {
            return false;
        }
        if ((m_wDate == null) || (W_StartDate == null) || m_wDate.before(W_StartDate)) {
            return false;
        }
        if ((m_wDate == null) || (W_EndDate == null) || m_wDate.after(W_EndDate)) {
            return false;
        }
        return true;
    }

    private final Task m_wTask;
    private final User m_wUser;
    private final Date m_wDate;
    private final double m_dImputed;

}
