/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Revision model class. */
class Revision extends Commentable {

    /** Constructor.
        @param W_Task           Task being revised.
        @param W_Date           Date of the revision.
        @param D_Remaining      Revision of the remaining value. Applicable only if positive or nul.
        @param D_Risks          Revision of the risks value. Applicable only if positive or nul.
        @param D_Opportunities  Revision of the opportunities value. Applicable only if positive or nul.
        @param D_Earned         Earned value. Applicable only if positive or nul. */
    public Revision(Task W_Task, Date W_Date, double D_Remaining, double D_Risks, double D_Opportunities, double D_Earned) {
        m_wTask = W_Task;
        m_wDate = new Date(W_Date);
        m_dRemaining = D_Remaining;
        m_dRisks = D_Risks;
        m_dOpportunities = D_Opportunities;
        m_dEarned = D_Earned;
    }

    /** String representation of the revision.
        @return The string representation of the revision. */
    public String toString() {
        return (
            "Revision "
            + "task = " + m_wTask + ", "
            + "date = " + m_wDate + ", "
            + "remaining = " + new Double(m_dRemaining) + ", "
            + "risks = " + new Double(m_dRisks) + ", "
            + "opportunities = " + new Double(m_dOpportunities) + ", "
            + "earned = " + new Double(m_dEarned)
        );
    }

    /** Revision merge operation.
        @param W_Contrib Revision contribution. */
    public boolean merge(Revision W_Contrib) {
        boolean b_Res = true;
        if (W_Contrib.hasRemaining()) {
            if (this.hasRemaining() && (this.getRemaining() != W_Contrib.getRemaining())) {
                b_Res = false;
            } else {
                m_dRemaining = W_Contrib.getRemaining();
            }
        }
        if (W_Contrib.hasRisks()) {
            if (this.hasRisks() && (this.getRisks() != W_Contrib.getRisks())) {
                b_Res = false;
            } else {
                m_dRisks = W_Contrib.getRisks();
            }
        }
        if (W_Contrib.hasOpportunities()) {
            if (this.hasOpportunities() && (this.getOpportunities() != W_Contrib.getOpportunities())) {
                b_Res = false;
            } else {
                m_dOpportunities = W_Contrib.getOpportunities();
            }
        }
        if (W_Contrib.hasEarned()) {
            if (this.hasEarned() && (this.getEarned() != W_Contrib.getEarned())) {
                b_Res = false;
            } else {
                m_dEarned = W_Contrib.getEarned();
            }
        }
        return b_Res;
    }

    /** Revised task accessor.
        @return The revised task. */
    public Task getTask() {
        return m_wTask;
    }

    /** Date of revision accessor.
        @return The date of the revision. */
    public Date getDate() {
        return new Date(m_wDate);
    }

    /** Remaining value validity predicate.
        @return true if the remain value should be taken in account, false if not. */
    public boolean hasRemaining() {
        return (m_dRemaining >= 0);
    }
    /** Remaining value accessor.
        @return The remaining value revision. */
    public double getRemaining() {
        return m_dRemaining;
    }

    /** Risks value validity predicate.
        @return true if the risks value should be taken in account, false if not. */
    public boolean hasRisks() {
        return (m_dRisks >= 0);
    }
    /** Risks value accessor.
        @return The risks value revision. */
    public double getRisks() {
        return m_dRisks;
    }

    /** Opportunities value validity predicate.
        @return true if the opportunities value should be taken in account, false if not. */
    public boolean hasOpportunities() {
        return (m_dOpportunities >= 0);
    }
    /** Opportunities value accessor.
        @return The opportunities value revision. */
    public double getOpportunities() {
        return m_dOpportunities;
    }

    /** Earned value validity predicate.
        @return true if the earned value should be taken in account, false if not. */
    public boolean hasEarned() {
        return (m_dEarned >= 0);
    }
    /** Earned value accessor.
        @return The earned value revision. */
    public double getEarned() {
        return m_dEarned;
    }

    private final Task m_wTask;
    private final Date m_wDate;
    private double m_dRemaining;
    private double m_dRisks;
    private double m_dOpportunities;
    private double m_dEarned;

}
