/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


import java.util.Vector;
import java.util.Iterator;
import java.util.Map;
import java.util.TreeMap;
import java.util.Set;


/** Task model class. */
class Task extends Commentable {

    /** No scheduling.
        The BCWS of the task itself will not be computed.
        Does not prevent BCWS computing for the sub-tasks. */
    public static final int SCHEDULING_NONE = 0;
    /** "At-end" scheduling.
        The BCWS of the task will be considered to be done at the end date only.
        Fits for development tasks of 15-20 days or less, which advancement is 0% until it is fully completed (100%). */
    public static final int SCHEDULING_AT_END = 1;
    /** Linear scheduling.
        The BCWS of the task will be approximated linearly from the planned starting date to the planned ending date of the task. */
    public static final int SCHEDULING_LINEAR = 2;

    /** Constructor.
        @param J_TaskId Identifier of the task.
        @param D_Reference              Reference value of the task. If negative, the task should not be checked against this value.
        @param D_ReferenceRisks         Reference risks value of the task.
        @param D_ReferenceOpportunities Reference opportunities value of the task.
        @param W_ReferenceStartDate     Reference starting date of the task.
                                        Can be null.
                                        If set, the reference value takes effect from this start date only.
                                        If set, imputations and revisions cannot be applied before that date neither.
        @param W_PlannedStartDate       Planned starting date of the task.
                                        Can be null.
                                        Useful with W_PlannedEndDate for BCWS computation.
        @param W_PlannedEndDate         Planned ending date of the task.
                                        Can be null.
                                        Useful with W_PlannedStartDate for BCWS computation.
        @param E_SchedulingMode         Scheduling mode. Either SCHEDULING_NONE, SCHEDULING_AT_END or SCHEDULING_LINEAR. */
    public Task(
            String J_TaskId,
            Date W_ReferenceStartDate, double D_Reference, double D_ReferenceRisks, double D_ReferenceOpportunities,
            Date W_PlannedStartDate, Date W_PlannedEndDate, int E_SchedulingMode) {

        // Initial value members.
        m_jTaskId = new String(J_TaskId);
        m_wParent = null;
        m_dReference = D_Reference;
        m_dReferenceRisks = D_ReferenceRisks;
        m_dReferenceOpportunities = D_ReferenceOpportunities;
        if (W_ReferenceStartDate != null) {
            m_wReferenceStartDate = new Date(W_ReferenceStartDate);
        } else {
            m_wReferenceStartDate = null;
        }
        if ((W_PlannedStartDate != null) && (W_PlannedEndDate != null)) {
            m_wPlannedStartDate = new Date(W_PlannedStartDate);
            m_wPlannedEndDate = new Date(W_PlannedEndDate);
        } else {
            m_wPlannedStartDate = null;
            m_wPlannedEndDate = null;
        }
        m_eSchedulingMode = E_SchedulingMode;

        // Computation members.
        m_jSubTasks = new Vector<Task>();
        m_jImputations = new TreeMap<Date, Vector<Imputation>>();
        m_jRevisions = new TreeMap<Date, Revision>();
        m_jComputations = new TreeMap<Date, Computation>();
        m_wComputedStartDate = null;
        m_wComputedEndDate = null;
        m_wLockEndDate = null;

        // Preliminary computations.
        // Set dummy revisions at the key dates, to force computation considering these dates.
        // Computed start and end dates will be adjusted by the way.
        if (m_wReferenceStartDate != null) {
            addRevision(new Revision(this, m_wReferenceStartDate, -1.0, -1.0, -1.0, -1.0));
        }
        if ((m_wPlannedStartDate != null) && (m_wPlannedEndDate != null)) {
            addRevision(new Revision(this, m_wPlannedStartDate, -1.0, -1.0, -1.0, -1.0));
            addRevision(new Revision(this, m_wPlannedEndDate, -1.0, -1.0, -1.0, -1.0));
        }
    }

    // Identifier.
    /** Identifier accessor.
        @return The identifier of the task. */
    public String getId() {
        return new String(m_jTaskId);
    }
    /** String representation of the task.
        @return The string representation of the task. */
    public String toString() {
        return ("Task '" + getId() + "'");
    }

    // Initial values.
    /** Reference value accessor.
        @return The reference value of the task. */
    public double getReference() {
        return m_dReference;
    }
    /** Reference risks value accessor.
        @return The reference risks value of the task. */
    public double getReferenceRisks() {
        return m_dReferenceRisks;
    }
    /** Reference opportunities value accessor.
        @return The reference opportunities value of the task. */
    public double getReferenceOpportunities() {
        return m_dReferenceOpportunities;
    }

    // Parent task / sub-tasks.
    /** Task/sub-task relation management.
        Permits the parent task positionning.

        When the start date of the task is not set, and the parent task has a start date set,
        the sub-task automatically inherits from this start date.

        @param W_Parent Parent task reference. */
    public void setParent(Task W_Parent) {
        if (W_Parent != null) {
            m_wParent = W_Parent;
            // If not set, inherit from the parent task reference start date.
            if (m_wReferenceStartDate == null) {
                if (m_wParent.getReferenceStartDate() != null) {
                    m_wReferenceStartDate = new Date(m_wParent.getReferenceStartDate());
                    addRevision(new Revision(this, m_wReferenceStartDate, -1.0, -1.0, -1.0, -1.0));
                }
            }
        }
    }
    /** Parent task accessor.
        @return The parent task reference. */
    public Task getParent() {
        return m_wParent;
    }
    /** Final characteristic accessor.
        Determines whether this task is final, ie it has no sub-task.
        @return true for final task, false otherwise. */
    public boolean isFinal() {
        return (! getSubTaskIterator().hasNext());
    }
    /** Sub-tasks iterator accessor.
        @return An iterator over the sub-tasks. */
    public Iterator<Task> getSubTaskIterator() {
        return m_jSubTasks.iterator();
    }
    /** Sub-task addition.
        @param W_Task Sub-task added. */
    public void addSubTask(Task W_Task) {
        if (W_Task != null) {
            W_Task.setParent(this);
            m_jSubTasks.add(W_Task);
            // Check the reference start date regarding the sub-task's one.
            if (    (m_wReferenceStartDate != null) && (W_Task.getReferenceStartDate() != null)
                    && m_wReferenceStartDate.after(W_Task.getReferenceStartDate())) {
                Display.warn("Incoherent reference start dates between "
                    + "the parent task '" + this.getId() + "' starting on " + this.getReferenceStartDate() + " "
                    + "and the sub-task '" + W_Task.getId() + "' starting on " + W_Task.getReferenceStartDate() + "."
                );
            }
        }
    }
    /** Task or sub-task research.
        Recursive search: looks in directs and deeper sub-tasks for the identifier researched.
        The research may also return the task itself if convenient.
        @param J_TaskId Identifier of the task researched. */
    public Task findTask(String J_TaskId) {
        if (J_TaskId != null) {
            // Look in sub-tasks before.
            if (m_jSubTasks != null) {
                for (   Iterator<Task> it = m_jSubTasks.iterator();
                        it.hasNext(); ) {
                    Task w_SubTask = it.next();
                    if (w_SubTask != null) {
                        Task w_Search = w_SubTask.findTask(J_TaskId);
                        if (w_Search != null) {
                            return w_Search;
                        }
                    }
                }
            }
            // Eventually reference for that task itself.
            if (m_jTaskId != null) {
                if (J_TaskId.equals(m_jTaskId)) {
                    return this;
                }
            }
        }
        return null;
    }

    // Imputations.
    /** Imputation addition.
        The imputations are stored in a collection member object, keeping them ready for later computations.
        @param W_Imputation Imputation added. */
    public void addImputation(Imputation W_Imputation) {
        if ((m_jImputations != null) && (W_Imputation != null)) {
            // Check the imputation does not come after the lock operation.
            if ((getLockEndDate() == null) || (! W_Imputation.getDate().after(getLockEndDate()))) {
                Vector<Imputation> j_ImputationsOfTheDate = null;
                if (m_jImputations.containsKey(W_Imputation.getDate())) {
                    j_ImputationsOfTheDate = m_jImputations.get(W_Imputation.getDate());
                } else {
                    j_ImputationsOfTheDate = new Vector<Imputation>();
                    m_jImputations.put(W_Imputation.getDate(), j_ImputationsOfTheDate);
                }
                if (j_ImputationsOfTheDate != null) {
                    j_ImputationsOfTheDate.add(W_Imputation);
                    checkBoundDates(W_Imputation.getDate());
                }
            }
        }
    }

    // Task revisions.
    /** Revision addition.
        The revisions are stored in a collection member object, keeping them ready for later computations.
        @param W_Revision Revision added. */
    public void addRevision(Revision W_Revision) {
        if ((W_Revision != null) && (m_jRevisions != null)) {
            // Check the revision does not come after the lock operation.
            if ((getLockEndDate() == null) || (! W_Revision.getDate().after(getLockEndDate()))) {
                Revision w_Revision = m_jRevisions.get(W_Revision.getDate());
                if (w_Revision != null) {
                    // If a revision already exists, merge the two elements.
                    if (! w_Revision.merge(W_Revision)) {
                        Display.warn("Merge failed between revisions " + this + " and " + W_Revision + ".");
                    }
                } else {
                    // Simple revision addition.
                    m_jRevisions.put(W_Revision.getDate(), W_Revision);
                    checkBoundDates(W_Revision.getDate());
                }
            }
        }
    }

    // Task lock.
    /** Lock attribute accessor.
        @return true when the task is locked, false otherwise. */
    public boolean isLocked() {
        return (m_wLockEndDate != null);
    }
    /** Lock operation.
        Sets the lock date for the task. Imputations and revisions after this date are then forbidden.
        This lock operation scans the imputations and revisions already attached to the task, and possibly makes corrections in coherence whith that principle.
        @param W_LockDate Lock date. */
    public void lock(Date W_LockDate) {
        if (isLocked()) {
            Display.warn("Task '" + getId() + "' is already locked since " + getLockEndDate() + ".");
        } else {
            Vector<Date> j_DatesToRemove = new Vector<Date>();

            // Check no imputation dates after the lock date.
            Set<Map.Entry<Date, Vector<Imputation>>> j_ImputationSet = m_jImputations.entrySet();
            for (   Iterator<Map.Entry<Date, Vector<Imputation>>> j_ImputationIt = j_ImputationSet.iterator();
                    j_ImputationIt.hasNext();
                    ) {
                Map.Entry<Date, Vector<Imputation>> j_ImputationEntry = j_ImputationIt.next();
                if ((j_ImputationEntry != null) && (j_ImputationEntry.getKey() != null) && (j_ImputationEntry.getValue() != null)) {
                    if (j_ImputationEntry.getKey().after(W_LockDate)) {
                        for (   Iterator<Imputation> it = j_ImputationEntry.getValue().iterator();
                                it.hasNext();
                                ) {
                            Imputation w_Imputation = it.next();
                            if (w_Imputation != null) {
                                Display.warn(w_Imputation + " is after the lock date " + W_LockDate + ". Removing imputation.");
                            }
                        }
                        j_DatesToRemove.add(j_ImputationEntry.getKey());
                    }
                }
            }

            // Check no revision dates after the lock date.
            Set<Map.Entry<Date, Revision>> j_RevisionSet = m_jRevisions.entrySet();
            for (   Iterator<Map.Entry<Date, Revision>> j_RevisionIt = j_RevisionSet.iterator();
                    j_RevisionIt.hasNext();
                    ) {
                Map.Entry<Date, Revision> j_RevisionEntry = j_RevisionIt.next();
                if ((j_RevisionEntry != null) && (j_RevisionEntry.getKey() != null) && (j_RevisionEntry.getValue() != null)) {
                    if (j_RevisionEntry.getKey().after(W_LockDate)) {
                        Revision w_Revision = j_RevisionEntry.getValue();
                        if (    (w_Revision != null)
                                // Display warnings for significant revisions only.
                                && (w_Revision.hasRemaining() || w_Revision.hasRisks() || w_Revision.hasOpportunities() || w_Revision.hasEarned())) {
                            Display.warn(w_Revision + " is after the lock date " + W_LockDate + ". Removing revision.");
                        }
                        j_DatesToRemove.add(j_RevisionEntry.getKey());
                    }
                }
            }

            for (   Iterator<Date> j_DateIt = j_DatesToRemove.iterator();
                    j_DateIt.hasNext();
                    ) {
                Date w_Date = j_DateIt.next();
                m_jImputations.remove(w_Date);
                m_jRevisions.remove(w_Date);
                // Remember however the date with a fake revision object, for computation reasons.
                m_jRevisions.put(w_Date, new Revision(this, w_Date, -1.0, -1.0, -1.0, -1.0));
            }

            // Propagate to sub-tasks.
            for (   Iterator<Task> it = m_jSubTasks.iterator();
                    it.hasNext(); ) {
                Task w_SubTask = it.next();
                if (w_SubTask != null) {
                    if (w_SubTask.isLocked()) {
                        if (W_LockDate.before(w_SubTask.getLockEndDate())) {
                            Display.warn("Incoherent lock date between "
                                + "the parent task '" + this.getId() + "' locked on " + W_LockDate + " "
                                + "and the sub-task '" + w_SubTask.getId() + "' locked on " + w_SubTask.getLockEndDate() + "."
                            );
                        }
                    } else {
                        w_SubTask.lock(W_LockDate);
                    }
                }
            }
            // Let's add a final revision to 0.
            // Consider the thing is done, the earned value reaches completion.
            addRevision(new Revision(this, W_LockDate, 0.0, 0.0, 0.0, getReference()));
            // Eventually set the lock date.
            m_wLockEndDate = new Date(W_LockDate);
        }
    }

    // Computation.
    /** Recursive computation method.
        The computation is disptached recursively on the tasks and sub-tasks. This recursive method is the one allowing such computation.
        @param W_Users List of users to focus the computation onto.
        @param W_ReferenceDate Reference date for the computation.
        @param W_StartDate Start date of the computation.
        @param W_EndDate End date of the computation.
        @param J_TraceOffset Useful for recursion debugging only.
        @see watch.Watch#execute */
    public void compute(
            User.Map W_Users, Date W_ReferenceDate, Date W_StartDate, Date W_EndDate,
            String J_TraceOffset
    ) {
        // Set dummy revisions for start and end dates of computation, so that the earn-value computation behaves correctly.
        if ((getPlannedStartDate() != null) && (getPlannedEndDate() != null)) {
            if (    (W_StartDate != null) && (! W_StartDate.equals(Date.zero()))
                    && (! W_StartDate.before(getPlannedStartDate())) && (! W_StartDate.after(getPlannedEndDate()))) {
                addRevision(new Revision(this, W_StartDate, -1.0, -1.0, -1.0, -1.0));
            }
            if (    (W_EndDate != null)
                    && (! W_EndDate.before(getPlannedStartDate())) && (! W_EndDate.after(getPlannedEndDate()))) {
                addRevision(new Revision(this, W_EndDate, -1.0, -1.0, -1.0, -1.0));
            }
        }

        // Check parameters.
        //if (W_StartDate == null) { W_StartDate = m_wReferenceStartDate; }     // Use the reference start date if it does exist.
        if (W_StartDate == null) { W_StartDate = Date.zero(); }
        if (W_EndDate == null) { W_EndDate = Date.today(); }
        if (W_ReferenceDate == null) {
            if ((m_wReferenceStartDate != null) && (m_wReferenceStartDate.after(Date.zero()))) {    // Use the reference start date if it does exist.
                W_ReferenceDate = new Date(m_wReferenceStartDate);
            } else {
                W_ReferenceDate = new Date(W_StartDate);
            }
        }

        String j_TraceStart = J_TraceOffset + "[" + getId() + "] ";

        Traces.traceComputation(j_TraceStart + "Task computation, task = " + this + "...");
        Traces.traceComputation(j_TraceStart + "W_Users = " + W_Users);
        Traces.traceComputation(j_TraceStart + "W_StartDate = " + W_StartDate);
        Traces.traceComputation(j_TraceStart + "W_ReferenceDate = " + W_ReferenceDate);
        Traces.traceComputation(j_TraceStart + "W_EndDate = " + W_EndDate);

        // Reset computations.
        m_jComputations.clear();

        // Initial values.
        double d_Reference = m_dReference;
        double d_ReferenceRisks = m_dReferenceRisks;
        double d_ReferenceOpportunities = m_dReferenceOpportunities;
        double d_Imputed = 0;
        double d_Remaining = m_dReference;
        double d_RemainingRisks = m_dReferenceRisks;
        double d_RemainingOpportunities = m_dReferenceOpportunities;
        double d_Planned = 0;
        double d_Earned = 0;
        Date w_SetReference = null;
        if ((m_wReferenceStartDate != null) && m_wReferenceStartDate.after(Date.zero())) {
            // Projects with initial start dates.
            d_Reference = 0;
            d_ReferenceRisks = 0;
            d_ReferenceOpportunities = 0;
            d_Remaining = 0;
            d_RemainingRisks = 0;
            d_RemainingOpportunities = 0;
            d_Planned = 0;
            d_Earned = 0;
            w_SetReference = new Date(m_wReferenceStartDate);
            if ((w_SetReference != null) && (w_SetReference.after(W_EndDate))) {
                // No reference at all for this task.
                w_SetReference = null;
            }
        }
        if (m_dReference < 0) {
            // Non-tracking tasks.
            d_Remaining = 0;
        }
        Date w_LastComputationDate = null;

        Computation w_InitialValues = new Computation(
            Date.zero(),
            d_Reference, d_ReferenceRisks, d_ReferenceOpportunities,
            d_Imputed,
            d_Remaining, d_RemainingRisks, d_RemainingOpportunities,
            d_Planned, d_Earned
        );
        Traces.traceComputation(j_TraceStart + "w_InitialValues = " + w_InitialValues);
        m_jComputations.put(Date.zero(), w_InitialValues);
        w_LastComputationDate = Date.zero();

        // Prepare revisions iteration.
        Set<Map.Entry<Date, Revision>> j_RevisionSet = m_jRevisions.entrySet();
        Iterator<Map.Entry<Date, Revision>> j_RevisionIt = j_RevisionSet.iterator();
        Revision w_Revision = null;
        if (j_RevisionIt.hasNext()) {
            w_Revision = j_RevisionIt.next().getValue();
        }

        // Prepare imputations iteration.
        Set<Map.Entry<Date, Vector<Imputation>>> j_ImputationSet = m_jImputations.entrySet();
        Iterator<Map.Entry<Date, Vector<Imputation>>> j_ImputationIt = j_ImputationSet.iterator();
        Map.Entry<Date, Vector<Imputation>> j_ImputationEntry = null;
        if (j_ImputationIt.hasNext()) {
            j_ImputationEntry = j_ImputationIt.next();
        }

        while ((w_Revision != null) || (j_ImputationEntry != null))
        {
            // Correct out of bounds items.
            while ( (w_Revision != null)
                    && (w_Revision.getDate().before(W_StartDate))) {
                if (j_RevisionIt.hasNext()) {
                    w_Revision = j_RevisionIt.next().getValue();
                } else {
                    w_Revision = null;
                }
            }
            if ((w_Revision != null) && (w_Revision.getDate().after(W_EndDate))) {
                w_Revision = null;
            }
            while ( (j_ImputationEntry != null)
                    && (j_ImputationEntry.getKey().before(W_StartDate))) {
                if (j_ImputationIt.hasNext()) {
                    j_ImputationEntry = j_ImputationIt.next();
                } else {
                    j_ImputationEntry = null;
                }
            }
            if ((j_ImputationEntry != null) && (j_ImputationEntry.getKey().after(W_EndDate))) {
                j_ImputationEntry = null;
            }

            // Determine which of revisions or imputations should be processed first.
            boolean b_ProcessRevision = false;
            if (false) {
            } else if (w_Revision == null) {
                // No more revisions. Process imputations
                b_ProcessRevision = false;
            } else if (j_ImputationEntry == null) {
                // No more imputations. Process revisions.
                b_ProcessRevision = true;
            } else {
                Date w_ImputationDate = j_ImputationEntry.getKey();
                if (w_ImputationDate.after(w_Revision.getDate())) {
                    // Next imputations stricly after the current revision. Process the revision first.
                    b_ProcessRevision = true;
                } else {
                    // Next imputations before the current revision. Process imputations first.
                    b_ProcessRevision = false;
                }
            }

            // Current date.
            Date w_Date = Date.zero();
            if (false) {
            } else if ((b_ProcessRevision) && (w_Revision != null)) {
                w_Date = w_Revision.getDate();
            } else if ((! b_ProcessRevision) && (j_ImputationEntry != null)) {
                w_Date = j_ImputationEntry.getKey();
            }
            Traces.traceComputation(j_TraceStart + "w_Date = " + w_Date);

            // Tasks with start dates.
            if ((w_SetReference != null) && (w_Date.after(w_SetReference) || w_Date.equals(w_SetReference))) {
                d_Reference = m_dReference;
                d_ReferenceRisks = m_dReferenceRisks;
                d_ReferenceOpportunities = m_dReferenceOpportunities;
                d_Remaining = m_dReference - d_Imputed;
                d_RemainingRisks = m_dReferenceRisks;
                d_RemainingOpportunities = m_dReferenceOpportunities;
                d_Planned = 0;
                d_Earned = 0;

                // Store a special computation.
                Computation w_Computation = new Computation(
                    w_SetReference,
                    d_Reference, d_ReferenceRisks, d_ReferenceOpportunities,
                    d_Imputed,
                    d_Remaining, d_RemainingRisks, d_RemainingOpportunities,
                    d_Planned, d_Earned
                );
                Traces.traceComputation(j_TraceStart + "w_Computation = " + w_Computation);
                m_jComputations.put(w_SetReference, w_Computation);
                w_LastComputationDate = w_SetReference;

                w_SetReference = null;
            }

            // Planned values.
            switch (m_eSchedulingMode) {
            case SCHEDULING_NONE:
                break;
            case SCHEDULING_AT_END:
                if (    (m_wPlannedEndDate != null) && (d_Planned <= 0.0)
                        && (w_Date.after(m_wPlannedEndDate) || w_Date.equals(m_wPlannedEndDate))) {
                    d_Planned = m_dReference;
                }
                break;
            case SCHEDULING_LINEAR:
                if (    (m_wPlannedStartDate != null) && (m_wPlannedEndDate != null)
                        && (w_Date.after(m_wPlannedStartDate) || w_Date.equals(m_wPlannedStartDate))
                        && (w_Date.before(m_wPlannedEndDate) || w_Date.equals(m_wPlannedEndDate))) {
                    // Compute planned values for all dates between the last computation date and the one we consider.
                    Date w_Start = null;
                    if (w_LastComputationDate != null) {
                        w_Start = w_LastComputationDate.nextDay();  // The last computation date should already be done.
                    }
                    if ((w_LastComputationDate == null) || m_wPlannedStartDate.after(w_LastComputationDate)) {
                        w_Start = m_wPlannedStartDate;
                    }
                    for (   Date w_It = w_Start;
                            w_It.before(w_Date) || w_It.equals(w_Date);
                            w_It = w_It.nextDay()) {
                        double d_PlannedDays = (double) (Date.delay(m_wPlannedStartDate, m_wPlannedEndDate) + 1);
                        double d_PassedDays = (double) (Date.delay(m_wPlannedStartDate, w_It) + 1);
                        d_Planned = m_dReference * d_PassedDays / d_PlannedDays;

                        // Unless the last one wich will follow the regular process.
                        if (w_It.before(w_Date)) {
                            Computation w_Computation = new Computation(
                                w_It,
                                d_Reference, d_ReferenceRisks, d_ReferenceOpportunities,
                                d_Imputed,
                                d_Remaining, d_RemainingRisks, d_RemainingOpportunities,
                                d_Planned, d_Earned
                            );
                            m_jComputations.put(w_It, w_Computation);
                        }
                    }
                }
                break;
            default:
                Display.error("Unknown scheduling mode " + m_eSchedulingMode);
                break;
            }

            // Values integration.
            if (b_ProcessRevision) {
                // Process the revision.
                Traces.traceComputation(j_TraceStart + "w_Revision = " + w_Revision);
                if (w_Revision != null) {
                    // Accept the revision values.
                    if (w_Revision.hasRemaining()) {
                        d_Remaining = w_Revision.getRemaining();
                        Traces.traceComputation(j_TraceStart + "d_Remaining = " + new Double(d_Remaining));
                    }
                    if (w_Revision.hasRisks()) {
                        d_RemainingRisks = w_Revision.getRisks();
                        Traces.traceComputation(j_TraceStart + "d_RemainingRisks = " + new Double(d_RemainingRisks));
                    }
                    if (w_Revision.hasOpportunities()) {
                        d_RemainingOpportunities = w_Revision.getOpportunities();
                        Traces.traceComputation(j_TraceStart + "d_RemainingOpportunities = " + new Double(d_RemainingOpportunities));
                    }
                    if (w_Revision.hasEarned()) {
                        d_Earned = w_Revision.getEarned();
                        Traces.traceComputation(j_TraceStart + "d_Earned = " + new Double(d_Earned));
                    }

                    // Move to next revision.
                    if (j_RevisionIt.hasNext()) {
                        w_Revision = j_RevisionIt.next().getValue();
                    } else {
                        w_Revision = null;
                    }
                }
            } else {
                if (j_ImputationEntry != null) {
                    // Process imputations.
                    Vector<Imputation> j_ImputationsOfTheDay = j_ImputationEntry.getValue();
                    if (j_ImputationsOfTheDay != null) {
                        // Sum up imputations of the day.
                        double d_ImputationsOfTheDay = 0;
                        for (   Iterator<Imputation> j_ImputationsOfTheDayIt = j_ImputationsOfTheDay.iterator();
                                j_ImputationsOfTheDayIt.hasNext(); ) {
                            Imputation w_Imputation = j_ImputationsOfTheDayIt.next();
                            if (w_Imputation != null) {
                                if (w_Imputation.matches(W_Users, W_StartDate, W_EndDate)) {
                                    Traces.traceComputation(j_TraceStart + "w_Imputation = " + w_Imputation);
                                    d_ImputationsOfTheDay += w_Imputation.getImputed();
                                } else {
                                    Traces.traceComputation(j_TraceStart + "Non matching w_Imputation = " + w_Imputation);
                                }
                            }
                        }
                        Traces.traceComputation(j_TraceStart + "d_ImputationsOfTheDay = " + new Double(d_ImputationsOfTheDay));

                        // Sum of imputations.
                        d_Imputed += d_ImputationsOfTheDay;
                        Traces.traceComputation(j_TraceStart + "d_Imputed = " + new Double(d_Imputed));

                        // Remaining revision of the day.
                        d_Remaining -= d_ImputationsOfTheDay;
                        if (d_Remaining < 0) {
                            d_Remaining = 0;
                        }
                        Traces.traceComputation(j_TraceStart + "d_Remaining = " + new Double(d_Remaining));
                    }

                    // Move to next imputation entry.
                    if (j_ImputationIt.hasNext()) {
                        j_ImputationEntry = j_ImputationIt.next();
                    } else {
                        j_ImputationEntry = null;
                    }
                }
            }

            // Final task reference.
            if (    (d_Reference >= 0)  // Note: Let negative reference task (non-tracking task) values remain to negative values.
                    && (w_Date.before(W_ReferenceDate) || w_Date.equals(W_ReferenceDate))) {
                d_Reference = d_Imputed + d_Remaining;
                d_ReferenceRisks = d_RemainingRisks;
                d_ReferenceOpportunities = d_RemainingOpportunities;
                Traces.traceComputation(j_TraceStart + "Final task reference adjustment");
                Traces.traceComputation(j_TraceStart + "d_Reference = " + new Double(d_Reference));
                Traces.traceComputation(j_TraceStart + "d_ReferenceRisks = " + new Double(d_ReferenceRisks));
                Traces.traceComputation(j_TraceStart + "d_ReferenceOpportunities = " + new Double(d_ReferenceOpportunities));
            }

            // Store computation of the day.
            Computation w_Computation = new Computation(
                w_Date,
                d_Reference, d_ReferenceRisks, d_ReferenceOpportunities,
                d_Imputed,
                d_Remaining, d_RemainingRisks, d_RemainingOpportunities,
                d_Planned, d_Earned
            );
            Traces.traceComputation(j_TraceStart + "w_Computation = " + w_Computation);
            m_jComputations.put(w_Date, w_Computation);
            w_LastComputationDate = w_Date;
        }

        // Propagate computation.
        for (   Iterator<Task> it = m_jSubTasks.iterator();
                it.hasNext(); ) {
            Task w_SubTask = it.next();
            if (w_SubTask != null) {
                // Propagate computation.
                w_SubTask.compute(W_Users, W_ReferenceDate, W_StartDate, W_EndDate, J_TraceOffset + " ");
                // Check start date and end date
                if (    (w_SubTask.getComputedStartDate() != null)
                        && ((m_wComputedStartDate == null) || w_SubTask.getComputedStartDate().before(m_wComputedStartDate))) {
                    m_wComputedStartDate = new Date(w_SubTask.getComputedStartDate());
                    Traces.traceComputation(
                        j_TraceStart + "m_wComputedStartDate = " + m_wComputedStartDate
                        + " because of " + w_SubTask
                    );
                }
                if (    (w_SubTask.getComputedEndDate() != null)
                        && ((m_wComputedEndDate == null) || w_SubTask.getComputedEndDate().after(m_wComputedEndDate))) {
                    m_wComputedEndDate = new Date(w_SubTask.getComputedEndDate());
                    Traces.traceComputation(
                        j_TraceStart + "m_wComputedEndDate = " + m_wComputedEndDate
                        + " because of " + w_SubTask
                    );
                }
            }
        }
    }
    /** Last computation accessor.
        Retrieves the last computation object after a computation launched previously.
        @return The last computation.
        @see watch.Task#compute */
    public Computation getLastComputation() {
        if (m_jImputations != null) {
            java.util.Collection<Computation> j_Computations = m_jComputations.values();
            if (j_Computations != null) {
                Computation w_LastComputation = null;
                for (   Iterator<Computation> it = j_Computations.iterator();
                        it.hasNext(); ) {
                    w_LastComputation = it.next();
                }
                if (w_LastComputation != null) {
                    return w_LastComputation;
                }
            }
        }
        return null;
    }
    /** Get the (cumulative) computation of the task up to a given date.
        @param W_UpTo Computation up to a given date.
        @param B_Cumulative Cumulative computation or not over the sub-tasks.
        @return The computation.
        @return null if an error occured. */
    public Computation getComputation(Date W_UpTo, boolean B_Cumulative) {
        try {
            double d_Reference = 0;
            double d_ReferenceRisks = 0;
            double d_ReferenceOpportunities = 0;
            double d_Imputed = 0;
            double d_Remaining = 0;
            double d_RemainingRisks = 0;
            double d_RemainingOpportunities = 0;
            double d_Planned = 0;
            double d_Earned = 0;

            // Find the computation requested for this task.
            if (true) {
                Computation w_Computation = null;
                java.util.Collection<Computation> j_Computations = m_jComputations.values();
                for (   Iterator<Computation> it = j_Computations.iterator();
                        it.hasNext(); ) {
                    Computation w_Tmp = it.next();
                    if (w_Computation == null) {
                        w_Computation = w_Tmp;
                    } else if (
                            (w_Tmp != null)
                            && w_Tmp.getDate().after(w_Computation.getDate())
                            && (w_Tmp.getDate().before(W_UpTo) || w_Tmp.getDate().equals(W_UpTo))
                    ) {
                        w_Computation = w_Tmp;
                    }
                }
                if (w_Computation != null) {
                    d_Reference += w_Computation.getReference();
                    d_ReferenceRisks += w_Computation.getReferenceRisks();
                    d_ReferenceOpportunities += w_Computation.getReferenceOpportunities();
                    d_Imputed += w_Computation.getImputed();
                    d_Remaining += w_Computation.getRemaining();
                    d_RemainingRisks += w_Computation.getRemainingRisks();
                    d_RemainingOpportunities += w_Computation.getRemainingOpportunities();
                    d_Planned += w_Computation.getPlanned();
                    d_Earned += w_Computation.getEarned();
                }
            }

            // Sum up over sub-tasks.
            if (B_Cumulative) {
                for (   java.util.Iterator<Task> it = m_jSubTasks.iterator();
                        it.hasNext(); ) {
                    Task w_SubTask = it.next();
                    if (w_SubTask != null) {
                        // Recursive calls.
                        Computation w_ComputationSum = w_SubTask.getComputation(W_UpTo, B_Cumulative);
                        // Local sum.
                        if (    (w_ComputationSum != null)
                                // Do not integrate negative reference tasks...
                                && (w_ComputationSum.getReference() >= 0)
                        ) {
                            d_Reference += w_ComputationSum.getReference();
                            d_ReferenceRisks += w_ComputationSum.getReferenceRisks();
                            d_ReferenceOpportunities += w_ComputationSum.getReferenceOpportunities();
                            d_Imputed += w_ComputationSum.getImputed();
                            d_Remaining += w_ComputationSum.getRemaining();
                            d_RemainingRisks += w_ComputationSum.getRemainingRisks();
                            d_RemainingOpportunities += w_ComputationSum.getRemainingOpportunities();
                            d_Planned += w_ComputationSum.getPlanned();
                            d_Earned += w_ComputationSum.getEarned();
                        }
                    }
                }
            }

            // Final result.
            return new Computation(
                W_UpTo,
                d_Reference, d_ReferenceRisks, d_ReferenceOpportunities,
                d_Imputed,
                d_Remaining, d_RemainingRisks, d_RemainingOpportunities,
                d_Planned, d_Earned
            );

        } catch (Exception e) {
            Display.printStackTrace(e);
            return null;
        }
    }
    /** Check computed start and end date regarding the given date.
        Corrects m_wComputedStartDate and m_wComputedEndDate depending on the given date.
        @param W_Date Computation date to take in account. */
    private void checkBoundDates(Date W_Date) {
        if (W_Date != null) {
            if ((m_wComputedStartDate == null) || (m_wComputedStartDate.after(W_Date))) {
                m_wComputedStartDate = new Date(W_Date);
            }
            if ((m_wComputedEndDate == null) || (m_wComputedEndDate.before(W_Date))) {
                m_wComputedEndDate = new Date(W_Date);
            }
        }
    }
    /** Reference start date accessor.
        @return The reference start date. null if no reference start date is set for the task. */
    public Date getReferenceStartDate() {
        if ((m_wReferenceStartDate == null) || (m_wReferenceStartDate.equals(Date.zero()))) {
            return null;
        } else {
            return new Date(m_wReferenceStartDate);
        }
    }
    /** Planned start date accessor.
        @return The planned start date. null if no planned start date is set for the task. */
    public Date getPlannedStartDate() {
        if ((m_wPlannedStartDate == null) || (m_wPlannedStartDate.equals(Date.zero()))) {
            return null;
        } else {
            return new Date(m_wPlannedStartDate);
        }
    }
    /** Planned end date accessor.
        @return The planned end date. null if no planned end date is set for the task. */
    public Date getPlannedEndDate() {
        if ((m_wPlannedEndDate == null) || (m_wPlannedEndDate.equals(Date.zero()))) {
            return null;
        } else {
            return new Date(m_wPlannedEndDate);
        }
    }
    /** Scheduling mode accessor.
        @return SCHEDULING_NONE if not applicable, the configured scheduling mode otherwise. */
    public int getSchedulingMode() {
        if ((getPlannedStartDate() == null) || (getPlannedEndDate() == null)) {
            return watch.Task.SCHEDULING_NONE;
        } else {
            return m_eSchedulingMode;
        }
    }
    /** Computed start date accessor.
        @return The computed start date. null if no start date has been computed. */
    public Date getComputedStartDate() {
        if ((m_wComputedStartDate == null) || m_wComputedStartDate.equals(Date.zero())) {
            return null;
        } else {
            return new Date(m_wComputedStartDate);
        }
    }
    /** Computed end date accessor.
        @return The compted end date. null if no end date has been computed. */
    public Date getComputedEndDate() {
        if (    (m_wComputedEndDate == null)
                || (m_wComputedStartDate == null) || m_wComputedStartDate.equals(Date.zero())) { // If the computed start date is not a valide date, the end date is no more valid.
            return null;
        } else {
            return new Date(m_wComputedEndDate);
        }
    }
    /** Lock end date accessor.
        @return The lock end date. null if the task is not locked. */
    public Date getLockEndDate() {
        if (m_wLockEndDate == null) {
            return null;
        } else {
            return new Date(m_wLockEndDate);
        }
    }
    /** Create the XML computation result.
        @param W_ParentNode Parent node to place the task computation result into. Either the root XML node, or the parent task XML result node.
        @param B_SumImputationsOnly When true, limits the XML result at the sum of imputations.
        @return A computation object representing the final status of the task. */
    public Computation mkXmlResult(org.w3c.dom.Element W_ParentNode, boolean B_SumImputationsOnly) {
        try {
            // Xml resource creations.
            org.w3c.dom.Element w_TaskNode = W_ParentNode.getOwnerDocument().createElement("task");
            W_ParentNode.appendChild(w_TaskNode);

            // Recursive calls.
            Computation w_TaskLastComputation = getLastComputation();
            double d_Reference = w_TaskLastComputation.getReference();
            double d_ReferenceRisks = w_TaskLastComputation.getReferenceRisks();
            double d_ReferenceOpportunities = w_TaskLastComputation.getReferenceOpportunities();
            double d_Imputed = w_TaskLastComputation.getImputed();
            double d_Remaining = w_TaskLastComputation.getRemaining();
            double d_RemainingRisks = w_TaskLastComputation.getRemainingRisks();
            double d_RemainingOpportunities = w_TaskLastComputation.getRemainingOpportunities();
            double d_Planned = w_TaskLastComputation.getPlanned();
            double d_Earned = w_TaskLastComputation.getEarned();
            for (   java.util.Iterator<Task> it = m_jSubTasks.iterator();
                    it.hasNext(); ) {
                Task w_SubTask = it.next();
                if (w_SubTask != null) {
                    // Recursive calls.
                    Computation w_ComputationSum = w_SubTask.mkXmlResult(w_TaskNode, B_SumImputationsOnly);
                    // Local sum.
                    if (    // Do not integrate negative reference tasks...
                            (w_ComputationSum.getReference() >= 0)
                            // ...unless we are in a imputation summing operation
                            || (B_SumImputationsOnly)) {
                        d_Reference += w_ComputationSum.getReference();
                        d_ReferenceRisks += w_ComputationSum.getReferenceRisks();
                        d_ReferenceOpportunities += w_ComputationSum.getReferenceOpportunities();
                        d_Imputed += w_ComputationSum.getImputed();
                        d_Remaining += w_ComputationSum.getRemaining();
                        d_RemainingRisks += w_ComputationSum.getRemainingRisks();
                        d_RemainingOpportunities += w_ComputationSum.getRemainingOpportunities();
                        d_Planned += w_ComputationSum.getPlanned();
                        d_Earned += w_ComputationSum.getEarned();
                    }
                }
            }

            // Finish: computation XML attributes.
            Traces.traceComputation("Task result, "
                + "task = " + this + ", "
                + "reference = " + new Double(d_Reference) + ", "
                + "reference-risks = " + new Double(d_ReferenceRisks) + ", "
                + "reference-opportunities = " + new Double(d_ReferenceOpportunities) + ", "
                + "imputed = " + new Double(d_Imputed) + ", "
                + "remaining = " + new Double(d_Remaining) + ", "
                + "remaining-risks = " + new Double(d_RemainingRisks) + ", "
                + "remaining-opportunities = " + new Double(d_RemainingOpportunities) + ", "
                + "planned = " + new Double(d_Planned) + ", "
                + "earned = " + new Double(d_Earned)
            );
            w_TaskNode.setAttribute("id", getId());
            if (! B_SumImputationsOnly) {
                w_TaskNode.setAttribute("reference", new Double(d_Reference).toString());
                w_TaskNode.setAttribute("reference-risks", new Double(d_ReferenceRisks).toString());
                w_TaskNode.setAttribute("reference-opportunities", new Double(d_ReferenceOpportunities).toString());
            }
            w_TaskNode.setAttribute("imputed", new Double(d_Imputed).toString());
            if (! B_SumImputationsOnly) {
                w_TaskNode.setAttribute("remaining", new Double(d_Remaining).toString());
                w_TaskNode.setAttribute("remaining-risks", new Double(d_RemainingRisks).toString());
                w_TaskNode.setAttribute("remaining-opportunities", new Double(d_RemainingOpportunities).toString());
            }
            if (! B_SumImputationsOnly) {
                w_TaskNode.setAttribute("planned", new Double(d_Planned).toString());
                w_TaskNode.setAttribute("earned", new Double(d_Earned).toString());
            }

            // Return a computation image.
            return new Computation(
                Date.today(),
                d_Reference, d_ReferenceRisks, d_ReferenceOpportunities,
                d_Imputed,
                d_Remaining, d_RemainingRisks, d_RemainingOpportunities,
                d_Planned, d_Earned
            );
        } catch (Exception e) {
            Display.printStackTrace(e);
            return null;
        }
    }



    private final Vector<Task> m_jSubTasks;
    private final String m_jTaskId;
    private Task m_wParent;
    private Date m_wReferenceStartDate;
    private final double m_dReference;
    private final double m_dReferenceRisks;
    private final double m_dReferenceOpportunities;
    private final Date m_wPlannedStartDate;
    private final Date m_wPlannedEndDate;
    private final int m_eSchedulingMode;
    private final TreeMap<Date, Vector<Imputation>> m_jImputations;
    private final TreeMap<Date, Revision> m_jRevisions;
    private final TreeMap<Date, Computation> m_jComputations;
    private Date m_wComputedStartDate;
    private Date m_wComputedEndDate;
    private Date m_wLockEndDate;

}
