/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Abstract class defining a post-computation transformation.
    @see watch.Watch#execute */
public abstract class Transformation {

    /** Identity transformation.
        The XML output structure directly coming from the computation should be kept as is for output. */
    public static final Transformation IDENTITY = new IdentityTransformation();
    /** Short transformation.
        The output is short single line giving the essentials of the imputations: amount of imputations, with the part of 'out-of-scope'.
        This output may basically be displayed in the standard output for coherence controls, month by month for instance. */
    public static final Transformation SHORT = new ShortTransformation();
    /** Text transformation.
        The output tables are displayed in a text output.
        Can be displayed either in the standard output or in a text file. */
    public static final Transformation TEXT = new TextTransformation();
    /** HTML transformation.
        The output tables are displayed in an HTML output.
        This output may basically be directed to an HTML file so that web browsers can display it correctly afterwards. */
    public static final Transformation HTML = new HtmlTransformation();
    /** Timeplot transformation.
        This transformation is the one basically used for history computations.
        It is actually composed by two transformation:
            <ul><li>a first one creating an HTML file loading the graph libraries and so,</li>
                <li>and a second one creating the data to load in the graph.</li></ul>
        @see watch.Watch#execute */
    public static final Transformation TIMEPLOT = new TimeplotTransformation();

    /** Constructor. */
    public Transformation() {
    }

    /** Post-computation transformation handler.
        @param XML_Input XML result of the watch computation.
        @param J_OutputFileName Name of the file resulting of the transformation. null corresponds to the standard output.
        @return true if success, false otherwise. */
    public abstract boolean apply(org.w3c.dom.Document XML_Input, String J_OutputFileName);

    /** XSL stylesheet application.
        This method is a tool method for subclasses.
        @param XML_Input XML input structure. Basically the XML result of watch computation.
        @param J_XslFileName Path of the XSL stylesheet to apply.
        @param J_OutputFileName Name of the file resulting of the XSL transformation. null corresponds to the standard output.
        @return true if success, false otherwise. */
    protected final boolean applyXsl(
            org.w3c.dom.Document XML_Input,
            String J_XslFileName,
            String J_OutputFileName) {
        try {

            Traces.traceTransformation("Transformation.applyXsl("
                + "XML_Input = " + XML_Input + ", "
                + "J_XslFileName = " + J_XslFileName + ", "
                + "J_OutputFileName = " + J_OutputFileName + ")"
            );

            // Determine output location.
            java.io.File j_OutputFile = null;
            if (J_OutputFileName != null) {
                j_OutputFile = PathControl.outputFile(J_OutputFileName);
            }

            // XSL stylesheet.
            javax.xml.transform.TransformerFactory xsl_Factory = javax.xml.transform.TransformerFactory.newInstance();
            javax.xml.transform.Transformer xsl_Transformer = null;
            if (J_XslFileName != null) {
                javax.xml.transform.stream.StreamSource xsl_Template = new javax.xml.transform.stream.StreamSource(
                    PathControl.xslFile(J_XslFileName)
                );
                xsl_Transformer = xsl_Factory.newTransformer(xsl_Template);
                // Timeplot stylesheet configuration.
                if ((j_OutputFile != null) && J_XslFileName.equals("xml2timeplot_html.xsl")) {
                    // Find out the base name of the output file.
                    String j_AbsolutePath = j_OutputFile.getAbsolutePath();
                    String j_BaseName = j_AbsolutePath;
                    java.io.File j_ParentDirectory = j_OutputFile.getParentFile();
                    if (j_ParentDirectory != null) {
                        String j_AbsoluteDir = j_ParentDirectory.getAbsolutePath();
                        if (j_AbsolutePath.substring(0, j_AbsoluteDir.length()).equals(j_AbsoluteDir)) {
                            j_BaseName = j_AbsolutePath.substring(j_AbsoluteDir.length() + 1, j_AbsolutePath.length());
                        }
                    }
                    // Let the XSL stylesheet know the name of the data file attached to the HTML history file.
                    xsl_Transformer.setParameter(
                        "STR_DataFile",
                        j_BaseName + ".txt"
                    );
                }
            } else {
                // Identity transformation stylesheet.
                xsl_Transformer = xsl_Factory.newTransformer();
                xsl_Transformer.setOutputProperty(javax.xml.transform.OutputKeys.INDENT, "yes");
                xsl_Transformer.setOutputProperty(javax.xml.transform.OutputKeys.ENCODING, "ISO-8859-1");
            }

            // XML source.
            javax.xml.transform.dom.DOMSource xml_Source = new javax.xml.transform.dom.DOMSource(XML_Input);
            javax.xml.transform.stream.StreamResult xml_Result = null;
            if (j_OutputFile != null) {
                xml_Result = new javax.xml.transform.stream.StreamResult(
                    new java.io.FileOutputStream(j_OutputFile)
                );
            } else {
                xml_Result = new javax.xml.transform.stream.StreamResult(new cli.OutputDevice.OutputStream(Display.out()));
            }

            // XSL Transformation.
            xsl_Transformer.transform(xml_Source, xml_Result);

            return true;

        } catch (Exception e) {
            Display.printStackTrace(e);
            return false;
        }
    }

}
