/*
    Copyright (c) 2008-2010, Alexis Royer, http://alexis.royer.free.fr/Watch

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
        * Neither the name of the Watch project nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package watch;


/** Watch application class. */
public class Watch {

    /** Version of the application. */
    public static final String VERSION = "2.0";
    /** Author of the application. */
    public static final String AUTHOR = "Alexis Royer";
    /** Email of the author. */
    public static final String EMAIL = "alexis.royer@gmail.com";
    /** URL of the project. */
    public static final String URL = "http://alexis.royer.free.fr/Watch/";
    /** Copyright notice. */
    public static final String COPYRIGHT = "(c) 2008-2010";

    /** Specific Watch application argments. */
    private static class WatchArgs extends Args {
        private final Option m_wInteractive;
        public WatchArgs() {
            super("Watch", "Watch");
            m_wInteractive = addBooleanOption('i', "interactive", "[{-i,--interactive}]", "-i,--interactive: Do not leave execution (verbose implied)");
                //if (arj_InputFiles.length <= 0) b_Interactive = true;
                //if (b_Interactive) b_Verbose = true;
        }
        public boolean getbInteractive() {
            return m_wInteractive.getbValue();
        }
    }

    /** Main function of the Watch application.
        Analyze arguments and launches the CLI.
        @param ARJ_Args Arguments passed to the program. */
    public static void main(String[] ARJ_Args) {
        try {

            cli.Cli.load();

            WatchArgs w_Args = new WatchArgs();
            w_Args.parse(ARJ_Args);

            run(new Watch(), w_Args);

        } catch (Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Main function of the Watch applications.
        @param W_Watch Watch instance to run.
        @param W_Args Arguments passed to the program. */
    protected static void run(Watch W_Watch, Args W_Args) {
        try {
            // Set the singleton.
            m_wInstance = W_Watch;

            Traces.traceApp("Creating a new IOMux...");
            cli.IOMux cli_IOMux = new cli.IOMux();
            cli_IOMux.setOutput(cli.Shell.ALL_STREAMS, cli.IODevice.getStdOut());
            cli_IOMux.setOutput(cli.Shell.ERROR_STREAM, cli.IODevice.getStdErr());
            cli.Traces.setStream(cli.IODevice.getStdErr());

            Traces.traceApp("Creating a new CLI...");
            cli.Cli cli_Cli = new WatchCli();
            Traces.traceApp("Creating a new shell...");
            cli.Shell cli_Shell = new cli.Shell(cli_Cli);

            // Input / output settings.
            if (! W_Args.getInputFiles().isEmpty()) {
                for (java.util.Iterator<String> j_It = W_Args.getInputFiles().iterator(); j_It.hasNext(); ) {
                    String j_InputFileName = j_It.next();
                    if (j_InputFileName != null) {
                        if (PathControl.isXml(j_InputFileName)) {
                            Traces.traceApp("Pushing a load command for the XML file '" + j_InputFileName + "' to IOMux...");
                            cli_IOMux.addInput(new cli.SingleCommand("xml-load \"" + j_InputFileName + "\"", cli.OutputDevice.getNullDevice()));
                        } else {
                            Traces.traceApp("Pushing an input file device for '" + j_InputFileName + "' to IOMux...");
                            cli_IOMux.addInput(new cli.InputFileDevice(j_InputFileName, cli.OutputDevice.getNullDevice()));
                        }
                    }
                }
            }
            if (! W_Args.getbVerbose()) {
                Traces.traceApp("Setting IOMux Welcome stream to null...");
                cli_IOMux.setOutput(cli.Shell.WELCOME_STREAM, cli.OutputDevice.getNullDevice());
                Traces.traceApp("Setting IOMux Echo stream to null...");
                cli_IOMux.setOutput(cli.Shell.ECHO_STREAM, cli.OutputDevice.getNullDevice());
            }
            Traces.traceApp("Setting IOMux Prompt stream to null...");
            cli_IOMux.setOutput(cli.Shell.PROMPT_STREAM, cli.OutputDevice.getNullDevice());
            if (W_Args instanceof WatchArgs) {
                if (((WatchArgs) W_Args).getbInteractive()) {
                    Traces.traceApp("Pushing a console device to IOMux (all outputs redirected)...");
                    cli.Console cli_Console = new cli.Console();
                    cli_IOMux.setOutput(cli.Shell.ALL_STREAMS, cli_Console);
                    cli_IOMux.addInput(cli_Console);
                    cli.Traces.setStream(cli_Console);
                }
            }
            // Eventually indicate the CLI to quit
            cli_IOMux.addInput(new cli.SingleCommand("quit", cli_IOMux));

            // Run the shell.
            Traces.traceApp("Running shell...");
            cli_Shell.run(cli_IOMux);

            cli.Traces.unsetStream();

        } catch (Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** CLI instance accessor.
        @return The only one CLI instance. */
    public static cli.Cli getCli() {
        java.util.Collection<cli.Cli> cli_List = new java.util.Vector<cli.Cli>();
        if (cli_List != null) {
            if (cli.Cli.findFromName(cli_List, ".*") > 0) {
                java.util.Iterator<cli.Cli> it = cli_List.iterator();
                if (it.hasNext()) {
                    return it.next();
                }
            }
        }

        return null;
    }

    /** CLI shell instance accessor.
        @return The only one CLI shell instance. */
    public static cli.Shell getShell() {
        cli.Cli cli_Cli = Watch.getCli();
        if (cli_Cli != null) {
            cli.Shell cli_Shell = cli_Cli.getShell();
            if (cli_Shell != null) {
                return cli_Shell;
            }
        }

        return null;
    }

    /** Watch application singleton variable.
        This member is protected so that Cli2Xml can set this singleton member with a Cli2Xml instance immediately. */
    protected static Watch m_wInstance = null;
    /** Watch application instance accessor.
        @return The only one Watch application instance. */
    public static Watch getInstance() {
        if (m_wInstance == null) {
            m_wInstance = new Watch();
        }
        return m_wInstance;
    }


    /** Constructor. */
    public Watch() {
        m_jProjectName = null;
        m_wMainTask = new Task("main", null, 0.0, 0.0, 0.0, null, null, Task.SCHEDULING_NONE);
        m_wOutOfScopeImputations = new Task("out-of-scope", null, -1.0, 0.0, 0.0, null, null, Task.SCHEDULING_NONE);
        m_wMainTask.addSubTask(m_wOutOfScopeImputations);
        m_wCurrentTask = m_wMainTask;

        m_wCurrentUser = null;
        m_wCurrentRevisionDate = null;
        m_wUserMap = new User.Map();
    }

    /** Load an XML watch syntax file.
        @param J_XmlPath Path of the XML watch syntax file to load. */
    public void load(String J_XmlPath) {
        Traces.traceInput("Load, xml-path = " + J_XmlPath);
        //  WatchXml w_XmlReader = new WatchXml();
        //  if (! w_XmlReader.readFile(J_XmlPath, getInstance())) {
        //      Display.error("Failed to load '" + J_XmlPath + "'");
        //  }
        //  
        try {
            new XmlInput_1_0().xml2model(new java.io.File(J_XmlPath), getInstance());
        } catch (Exception j_Exception) {
            Display.printStackTrace(j_Exception);
        }
    }

    /** Comment input.
        @param J_Words Collection of words, making a sentence. */
    public void comment(java.util.Collection<String> J_Words) {
    }
    /** Comment input.
        @param J_Comment Comment sentence. */
    public final void comment(String J_Comment) {
        java.util.Collection<String> j_Words = new java.util.Vector<String>();
        j_Words.add(J_Comment);
        comment(j_Words);
    }

    /** Project name declaration.
        @param J_ProjectName New project name. */
    public void projectName(String J_ProjectName) {
        try {
            if (J_ProjectName != null) {
                Traces.traceInput("Project name, name = " + J_ProjectName);
                m_jProjectName = new String(J_ProjectName);
            } else {
                m_jProjectName = null;
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Locale declaration.
        @param J_Locale New locale in its string form:
                        <ul><li>'chinese'</li>
                            <li>'english'</li>
                            <li>'french'</li>
                            <li>'german'</li>
                            <li>'italian'</li>
                            <li>'japanese'.</li></ul> */
    public void locale(String J_Locale) {
        java.util.Locale j_Locale = null;
        if (false) {
        } else if (J_Locale.equalsIgnoreCase("chinese")) {
            j_Locale = java.util.Locale.CHINESE;
        } else if (J_Locale.equalsIgnoreCase("english")) {
            j_Locale = java.util.Locale.ENGLISH;
        } else if (J_Locale.equalsIgnoreCase("french")) {
            j_Locale = java.util.Locale.FRENCH;
        } else if (J_Locale.equalsIgnoreCase("german")) {
            j_Locale = java.util.Locale.GERMAN;
        } else if (J_Locale.equalsIgnoreCase("italian")) {
            j_Locale = java.util.Locale.ITALIAN;
        } else if (J_Locale.equalsIgnoreCase("japanese")) {
            j_Locale = java.util.Locale.JAPANESE;
        }

        if (j_Locale == null) {
            Display.error("Watch.locale(): Bad parameter!");
        } else {
            Traces.traceInput("locale = " + j_Locale);
            Date.setCurrentLocale(j_Locale);
        }
    }

    /** Beginning of a task group.
        @param J_TaskId Task identifier.
        @param J_StartDate Task start date. Can be null if no specific start date. */
    public void beginTaskGroup(String J_TaskId, String J_StartDate) {
        try {
            if (J_TaskId == null) {
                Display.error("Watch.beginTaskGroup(): Bad parameter!");
            } else {
                if (m_wCurrentTask == null) {
                    Display.error("Internal error!");
                } else {
                    Date w_StartDate = null;
                    if (J_StartDate != null) {
                        w_StartDate = new Date(J_StartDate);
                    }
                    Task w_NewTask = new Task(J_TaskId, w_StartDate, 0.0, 0.0, 0.0, null, null, Task.SCHEDULING_NONE);
                    if (w_NewTask == null) {
                        Display.error("Could not create a new task!");
                    } else {
                        Traces.traceInput(
                            "Beginning task group, parent task = " + m_wCurrentTask + ", new task = " + w_NewTask + ", "
                            + "start date = " + w_StartDate
                        );
                        m_wCurrentTask.addSubTask(w_NewTask);
                        m_wCurrentTask = w_NewTask;
                    }
                }
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** End of a task group. */
    public void endTaskGroup() {
        try {
            if (m_wCurrentTask == null) {
                Display.error("Internal error!");
            } else {
                Task w_ParentTask = m_wCurrentTask.getParent();
                if (w_ParentTask == null) {
                    Display.error("Already on the root task!");
                } else {
                    Traces.traceInput("Ending task group, ending task = " + m_wCurrentTask + ", parent task = " + w_ParentTask);
                    m_wCurrentTask = w_ParentTask;
                }
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Single task definition.
        @param J_TaskId                 Task identifier.
        @param J_ReferenceStartDate     Reference start date. Can be null if no specific reference start date.
        @param D_Reference              Reference value for the task. -1.0 if this task should not be checked against its reference.
        @param D_ReferenceRisks         Reference risks of the task.
        @param D_ReferenceOpportunities Reference opportunities of the task.
        @param J_PlannedStartDate       Planned start date. Can be null if no planification is done.
        @param J_PlannedEndDate         Planned end date. Can be null if no planification is done.
        @param E_SchedulingMode         Scheduling mode. Can be either Task.SCHEDULING_AT_END or Task.SCHEDULING_LINEAR. */
    public void singleTask(
            String J_TaskId,
            String J_ReferenceStartDate, double D_Reference, double D_ReferenceRisks, double D_ReferenceOpportunities,
            String J_PlannedStartDate, String J_PlannedEndDate, int E_SchedulingMode) {
        try {
            if (J_TaskId == null) {
                Display.error("Watch.singleTask(): Bad parameter!");
            } else {
                if (m_wCurrentTask == null) {
                    Display.error("Internal error!");
                } else {
                    Date w_ReferenceStartDate = null;
                        if (J_ReferenceStartDate != null) { w_ReferenceStartDate = new Date(J_ReferenceStartDate); }
                    Date w_PlannedStartDate = null;
                        if (J_PlannedStartDate != null) { w_PlannedStartDate = new Date(J_PlannedStartDate); }
                    Date w_PlannedEndDate = null;
                        if (J_PlannedEndDate != null) { w_PlannedEndDate = new Date(J_PlannedEndDate); }
                    if ((w_PlannedStartDate != null) && (w_PlannedEndDate != null) && (E_SchedulingMode == Task.SCHEDULING_NONE)) {
                        E_SchedulingMode = Task.SCHEDULING_AT_END;
                    }
                    Task w_NewTask = new Task(
                        J_TaskId,
                        w_ReferenceStartDate, D_Reference, D_ReferenceRisks, D_ReferenceOpportunities,
                        w_PlannedStartDate, w_PlannedEndDate, E_SchedulingMode
                    );
                    if (w_NewTask == null) {
                        Display.error("Could not create a new task!");
                    } else {
                        String j_SchedulingMode = null;
                        switch (E_SchedulingMode) {
                        case Task.SCHEDULING_AT_END:    j_SchedulingMode = "at-end scheduling";         break;
                        case Task.SCHEDULING_LINEAR:    j_SchedulingMode = "linear scheduling";         break;
                        default:                        j_SchedulingMode = "unknown scheduling mode";   break;
                        }
                        Traces.traceInput("Single task, parent task = " + m_wCurrentTask + ", single task = " + w_NewTask + ", "
                            + "value = " + new Double(w_NewTask.getReference()) + ", "
                            + "start date = " + w_NewTask.getReferenceStartDate() + ", "
                            + "risks = " + new Double(w_NewTask.getReferenceRisks()) + ", "
                            + "opportunities = " + new Double(w_NewTask.getReferenceOpportunities()) + ", "
                            + "planned from " + w_NewTask.getPlannedStartDate() + " to " + w_NewTask.getPlannedEndDate() + ", "
                            + j_SchedulingMode
                        );
                        m_wCurrentTask.addSubTask(w_NewTask);
                    }
                }
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Current user setting. Useful for imputations for instance.
        @param J_UserName User identifier: either the full name, or initials... whatever, it is up to the user to choose. */
    public void setCurrentUser(String J_UserName) {
        try {
            if (J_UserName == null) {
                Display.error("Watch.setCurrentUser(): Bad parameter!");
            } else {
                m_wCurrentUser = m_wUserMap.ensure(J_UserName);
                Traces.traceInput("Current user = " + m_wCurrentUser);
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Make an imputation.
        Within the context of the user defined by Watch.setCurrentUser().
        @param J_TaskId Identifier of the task the imputation is related to.
        @param J_Date Date of the imputation.
        @param D_Imputed Imputation value. */
    public void imputation(String J_TaskId, String J_Date, double D_Imputed) {
        try {
            if ((J_TaskId == null) || (J_Date == null)) {
                Display.error("Watch.imputation(): Bad parameter!");
            } else {
                Date w_ImputationDate = new Date(J_Date);
                if (m_wMainTask == null) {
                    Display.error("Internal error!");
                } else {
                    Task w_Task = m_wMainTask.findTask(J_TaskId);
                    if (w_Task == null) {
                        Display.warn(
                            m_wCurrentUser + ", "
                            + w_ImputationDate + ", "
                            + "out of scope task id '" + J_TaskId + "'"
                        );
                        w_Task = m_wOutOfScopeImputations;
                    }
                    if ((w_Task == null) || (m_wCurrentUser == null)) {
                        Display.error("Internal error!");
                    } else {
                        if (! w_Task.isFinal()) {
                            Display.warn("Non-final task '" + J_TaskId + "'. Ignoring imputation for " + m_wCurrentUser + " on " + w_ImputationDate + ".");
                        /*} else if ((w_Task.getReferenceStartDate() != null) && w_ImputationDate.before(w_Task.getReferenceStartDate())) {
                            Display.warn("Task '" + J_TaskId + "' starts on " + w_Task.getReferenceStartDate() + " only. Ignoring imputation for " + m_wCurrentUser + " on " + w_ImputationDate + ".");*/
                        } else if (w_Task.isLocked() && w_ImputationDate.after(w_Task.getLockEndDate())) {
                            Display.warn("Task '" + J_TaskId + "' is locked since " + w_Task.getLockEndDate() + ". Ignoring imputation for " + m_wCurrentUser + " on " + w_ImputationDate + ".");
                        } else {
                            Imputation w_Imputation = new Imputation(w_Task, m_wCurrentUser, w_ImputationDate, D_Imputed);
                            w_Task.addImputation(w_Imputation);
                            Traces.traceInput("new imputation = " + w_Imputation);
                        }
                    }
                }
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Current revision date setting. Permits Watch.revision() and Watch.lock() to be called without repeating the revision date.
        @param J_Date Revision date to remember for next revisions and lock operations. */
    public void setCurrentRevisionDate(String J_Date) {
        try {
            if (J_Date != null) {
                m_wCurrentRevisionDate = new Date(J_Date);
            } else {
                //Display.error("Watch.setCurrentRevisionDate(): Bad parameter!");
                m_wCurrentRevisionDate = null;
            }
        } catch (Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Make a revision.
        @param J_TaskId         Identifier of the task wich remaining value is being revised.
        @param J_Date           Date of the revision. Can be null if Watch.setCurrentDateRevision() has been called previously.
        @param D_Remaining      Revision of the remaining value. Applicable only if positive or nul.
        @param D_Risks          Revision of the risks value. Applicable only if positive or nul.
        @param D_Opportunities  Revision of the opportunities value. Applicable only if positive or nul.
        @param D_Earned         Revision of the earned value. Applicable only if positive or nul. */
    public void revision(String J_TaskId, String J_Date, double D_Remaining, double D_Risks, double D_Opportunities, double D_Earned) {
        try {
            if (J_TaskId == null) {
                Display.error("Watch.revision(): Bad parameter!");
            } else {
                if (m_wMainTask == null) {
                    Display.error("Internal error!");
                } else {
                    // Determine the revision date.
                    Date w_RevisionDate = m_wCurrentRevisionDate;   // Either the global one for the revision section.
                    if (J_Date != null) {
                        w_RevisionDate = new Date(J_Date);          // Or a specific one for this command.
                    }
                    if (w_RevisionDate == null) {
                        Display.error("Please give a revision date.");
                    } else {
                        // Find out the revised task.
                        Task w_Task = m_wMainTask.findTask(J_TaskId);
                        if (w_Task == null) {
                            Display.warn("No such task '" + J_TaskId + "'. Ignoring revision on " + w_RevisionDate + ".");
                        } else {
                            if (! w_Task.isFinal()) {
                                Display.warn("Non-final task '" + J_TaskId + "'. Ignoring revision on " + w_RevisionDate + ".");
                            /*} else if ((w_Task.getReferenceStartDate() != null) && w_RevisionDate.before(w_Task.getReferenceStartDate())) {
                                Display.warn("Task '" + J_TaskId + "' starts on " + w_Task.getReferenceStartDate() + " only. Ignoring revision on " + w_RevisionDate + ".");*/
                            } else if (w_Task.isLocked() && w_RevisionDate.after(w_Task.getLockEndDate())) {
                                Display.warn("Task '" + J_TaskId + "' is locked since " + w_Task.getLockEndDate() + ". Ignoring revision on " + w_RevisionDate + ".");
                            } else if ((D_Remaining < 0.0) && (D_Risks < 0.0) && (D_Opportunities < 0.0) && (D_Earned < 0.0)) {
                                Display.warn("Please revise the remaining, the risks, the opportunities or the earned value at least. Ignoring useless revision on " + w_RevisionDate + ".");
                            } else {
                                Revision w_Revision = new Revision(w_Task, w_RevisionDate, D_Remaining, D_Risks, D_Opportunities, D_Earned);
                                w_Task.addRevision(w_Revision);
                                Traces.traceInput("new revision = " + w_Revision);
                            }
                        }
                    }
                }
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Make a lock operation. The task locked cannot be imputed neither reviewed after the lock date.
        @param J_TaskId Identifier of the task being locked.
        @param J_Date Date of the lock operation. Can be null if Watch.setCurrentDateRevision() has been called previously. */
    public void lock(String J_TaskId, String J_Date) {
        try {
            if (J_TaskId == null) {
                Display.error("Watch.lock(): Bad parameter!");
            } else {
                if (m_wMainTask == null) {
                    Display.error("Internal error!");
                } else {
                    // Determine the revision date.
                    Date w_LockDate = m_wCurrentRevisionDate;   // Either the global one for the revision section.
                    if (J_Date != null) {
                        w_LockDate = new Date(J_Date);          // Or a specific one for this command.
                    }
                    if (w_LockDate == null) {
                        Display.error("Please give a lock or revision date.");
                    } else {
                        // Find out the revised task.
                        Task w_Task = m_wMainTask.findTask(J_TaskId);
                        if (w_Task == null) {
                            Display.warn("No such task '" + J_TaskId + "'. Ignoring revision.");
                        } else {
                            if (w_Task.isLocked()) {
                                Display.warn("Task '" + J_TaskId + "' is already locked since " + w_Task.getLockEndDate() + ". Ignoring lock operation on " + w_LockDate + ".");
                            /*} else if ((w_Task.getReferenceStartDate() != null) && w_LockDate.before(w_Task.getReferenceStartDate())) {
                                Display.warn("Task '" + J_TaskId + "' starts on " + w_Task.getReferenceStartDate() + " only. Ignoring lock operation on " + w_LockDate + ".");*/
                            } else {
                                w_Task.lock(w_LockDate);
                                Traces.traceInput("task " + w_Task + " locked on " + w_LockDate);
                            }
                        }
                    }
                }
            }
        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }

    /** Watch execution.
        Consists in a computation followed by an output, either in the standard output or in an output file.
        @param J_FocusTask          Task to focus the computation onto. If null, the whole project will be computed.
        @param J_Users              List of users to focus the computation onto. If null or empty, all users imputations will be taken in account.
        @param J_ReferenceDate      Date to consider as the reference date. This can allow one to check the project against a certain date.
                                    If null, the initial values remain the reference.
        @param J_StartDate          Date to start the computation from. Imputations and revisions before will not be taken in account.
                                    Useful if B_SumImputationsOnly is true.
                                    If null, no specific start date will be considered for imputations and revisions validity.
        @param J_EndDate            Date to end the computation at. Imputations and revisions after will not be taken in account.
                                    Useful if B_SumImputationsOnly is true.
                                    If null, no specific and date will be considered for imputations and revisions validity.
        @param B_SumImputationsOnly When true, limits the computation at summing imputations.
                                    Ignored when B_History is true.
        @param B_History            When true, launches the history computation instead of a regular computation.
                                    The history computation is different in the way that it retrieves reference, imputed, remaining, risks and opportunities values along the time for the focus task,
                                    whereas the regular computation retrieves these values for the focus task and all its sub-tasks, but at the end date only.
                                    The history computation implies an history transformation for W_Transformation.
        @param W_Transformation     The computation results in an XML structure in memory.
                                    This parameter indicates the transformation to apply for presenting the results.
        @param J_OutputFileName     Results output file name.
                                    If null, the output is the standard output. */
    public void execute(
            String J_FocusTask, java.util.Collection<String> J_Users,
            String J_ReferenceDate, String J_StartDate, String J_EndDate,
            boolean B_SumImputationsOnly, boolean B_History,
            Transformation W_Transformation,
            String J_OutputFileName) {

        try {
            User.Map w_Users = m_wUserMap.getListOf(J_Users);
            Traces.traceInput("watch execution: "
                + "J_FocusTask = " + J_FocusTask + ", "
                + "J_Users = " + w_Users + ", "
                + "J_ReferenceDate = " + J_ReferenceDate + ", "
                + "J_StartDate = " + J_StartDate + ", "
                + "J_EndDate = " + J_EndDate + ", "
                + "B_SumImputationsOnly = " + new Boolean(B_SumImputationsOnly).toString() + ", "
                + "B_History = " + new Boolean(B_History).toString() + ", "
                + "W_Transformation = " + W_Transformation + ", "
                + "J_OutputFileName = " + J_OutputFileName);

            if (J_OutputFileName != null) {
                Display.out().put("Watching ");
                if (J_FocusTask != null) {
                    Display.out().put("task '" + J_FocusTask + "' ");
                } else {
                    Display.out().put("project ");
                    if (m_jProjectName != null) {
                        Display.out().put("'" + m_jProjectName + "' ");
                    }
                }
                if (! w_Users.isEmpty()) {
                    Display.out().put("for users " + w_Users + " ");
                }
                if (J_StartDate != null) {
                    Display.out().put("from " + J_StartDate + " ");
                }
                if (J_EndDate != null) {
                    Display.out().put("up to " + J_EndDate + " ");
                }
                if (J_ReferenceDate != null) {
                    Display.out().put("refering to " + J_ReferenceDate + " ");
                }
                if (J_OutputFileName != null) {
                    Display.out().put("to file " + J_OutputFileName + " ");
                }
                if (B_SumImputationsOnly) {
                    Display.out().put("(sum) ");
                } else if (B_History) {
                    Display.out().put("(history) ");
                }
                Display.out().endl();
            }

            Task w_FocusTask = m_wMainTask;
            if ((J_FocusTask != null) && (m_wMainTask != null)) {
                w_FocusTask = m_wMainTask.findTask(J_FocusTask);
                if (w_FocusTask == null) {
                    w_FocusTask = m_wMainTask;
                    Display.warn(
                        "No such task '" + J_FocusTask + "'. "
                        + "Using main task instead."
                    );
                }
            }

            if (W_Transformation == null) {
                Display.error("Watch.execute(): Bad parameter!");
            } else {
                if (w_FocusTask == null) {
                    Display.error("Internal error!");
                } else {
                    Traces.traceComputation("Starting computation...");

                    Date w_ReferenceDate = Date.zero();
                    if (J_ReferenceDate != null) {
                        w_ReferenceDate = new Date(J_ReferenceDate);
                    }
                    Date w_StartDate = Date.zero();
                    if (J_StartDate != null) {
                        w_StartDate = new Date(J_StartDate);
                    }
                    Date w_EndDate = Date.today();
                    if (J_EndDate != null) {
                        w_EndDate = new Date(J_EndDate);
                    }

                    org.w3c.dom.Document xml_Doc = null;
                    if (B_History) {
                        xml_Doc = history(
                            w_FocusTask, w_Users,
                            w_ReferenceDate, w_StartDate, w_EndDate
                        );
                    } else {
                        xml_Doc = watch(
                            w_FocusTask, w_Users,
                            w_ReferenceDate, w_StartDate, w_EndDate,
                            B_SumImputationsOnly
                        );
                    }
                    Traces.traceTransformation("Watch.execute(): W_Transformation = " + W_Transformation + ", xml_Doc = " + xml_Doc + ", J_OutputFileName = " + J_OutputFileName);
                    W_Transformation.apply(xml_Doc, J_OutputFileName);
                }
            }

        } catch(Exception e) {
            Display.printStackTrace(e);
        }
    }
    private org.w3c.dom.Document watch(
            Task W_FocusTask, User.Map W_Users,
            Date W_ReferenceDate, Date W_StartDate, Date W_EndDate,
            boolean B_SumImputationsOnly
    ) {
        try {
            W_FocusTask.compute(W_Users, W_ReferenceDate, W_StartDate, W_EndDate, "");

            // Root document.
            javax.xml.parsers.DocumentBuilderFactory xml_BuilderFactory = javax.xml.parsers.DocumentBuilderFactory.newInstance();
            javax.xml.parsers.DocumentBuilder xml_Builder = xml_BuilderFactory.newDocumentBuilder();
            org.w3c.dom.Document xml_Doc = xml_Builder.newDocument();
            org.w3c.dom.Element xml_RootNode = xml_Doc.createElement("watch");
            xml_Doc.appendChild(xml_RootNode);
                // Project name.
            String j_ProjectName = m_jProjectName;
            if ((W_FocusTask != null) && (W_FocusTask != m_wMainTask)) {
                j_ProjectName = W_FocusTask.getId();
            }
            if (j_ProjectName != null) {
                xml_RootNode.setAttribute("project", j_ProjectName);
            }
                // User.
            if ((W_Users != null) && (! W_Users.isEmpty())) {
                xml_RootNode.setAttribute("user", W_Users.toString());
            }
                // Reference date.
            if (! B_SumImputationsOnly) {
                if ((W_ReferenceDate != null) && (! W_ReferenceDate.equals(Date.zero()))) {
                    xml_RootNode.setAttribute("reference", W_ReferenceDate.toString());
                }
            }
                // Start date.
            if ((W_StartDate != null) && (! W_StartDate.equals(Date.zero()))) {
                xml_RootNode.setAttribute("start", W_StartDate.toString());
            }
                // End date.
            if (W_EndDate != null) {
                xml_RootNode.setAttribute("end", W_EndDate.toString());
            }

            // Recursive calls.
            W_FocusTask.mkXmlResult(xml_RootNode, B_SumImputationsOnly);

            // Finish.
            return xml_Doc;
        } catch (Exception e) {
            Display.printStackTrace(e);
            return null;
        }
    }
    private org.w3c.dom.Document history(
            Task W_FocusTask, User.Map W_Users,
            Date W_ReferenceDate, Date W_StartDate, Date W_EndDate) {

        try {
            W_FocusTask.compute(W_Users, W_ReferenceDate, W_StartDate, W_EndDate, "");

            // Root document.
            javax.xml.parsers.DocumentBuilderFactory xml_BuilderFactory = javax.xml.parsers.DocumentBuilderFactory.newInstance();
            javax.xml.parsers.DocumentBuilder xml_Builder = xml_BuilderFactory.newDocumentBuilder();
            org.w3c.dom.Document xml_Doc = xml_Builder.newDocument();
            org.w3c.dom.Element xml_RootNode = xml_Doc.createElement("watch");
            xml_Doc.appendChild(xml_RootNode);
                // Project name.
            String j_ProjectName = m_jProjectName;
            if (    (W_FocusTask != null)
                    && ((j_ProjectName == null) || (W_FocusTask != m_wMainTask))) {
                j_ProjectName = W_FocusTask.getId();
            }
            if (j_ProjectName != null) {
                xml_RootNode.setAttribute("project", j_ProjectName);
            }
                // User.
            if (W_Users != null) {
                xml_RootNode.setAttribute("user", W_Users.toString());
            }
                // Reference date.
            if ((W_ReferenceDate != null) && (! W_ReferenceDate.equals(Date.zero()))) {
                xml_RootNode.setAttribute("reference", W_ReferenceDate.toString());
            }
                // Start date.
            if ((W_StartDate != null) && (! W_StartDate.equals(Date.zero()))) {
                xml_RootNode.setAttribute("start", W_StartDate.toString());
            }
                // End date.
            if (W_EndDate != null) {
                xml_RootNode.setAttribute("end", W_EndDate.toString());
            }
                // Date format.
            xml_RootNode.setAttribute("sample-date", new Date("2000-12-31").toString());

            // Determine start and end dates.
            Date w_StartDate = W_FocusTask.getComputedStartDate();
            Date w_EndDate = W_FocusTask.getComputedEndDate();
            if ((w_StartDate == null) || (w_EndDate == null)) {
                // In case of invalid start and/or end dates, show history between yesterday and today.
                w_EndDate = Date.today();
                w_StartDate = w_EndDate.prevDay();
            }
            if ((W_StartDate == null) || (W_StartDate.equals(Date.zero()))) {
                // If the requested start date is not specified,
                Computation w_InitialComputation = W_FocusTask.getComputation(Date.zero(), true);
                if (w_InitialComputation.getReference() != 0) {
                    // and the initial computation is significant enough,
                    // start with the day before.
                    w_StartDate = w_StartDate.prevDay();
                }
            }
            Traces.traceComputation("history start date = " + w_StartDate);
            Traces.traceComputation("history end date = " + w_EndDate);

            // Scan all dates, and make recursive calls.
            for (   Date w_DateIt = w_StartDate;
                    w_DateIt.before(w_EndDate) || w_DateIt.equals(w_EndDate);
                    w_DateIt = w_DateIt.nextDay()) {

                // Finish: computation XML attributes.
                Traces.traceComputation("History computation for " + w_DateIt);

                Computation w_Computation = W_FocusTask.getComputation(w_DateIt, true);
                if (w_Computation != null) {
                    org.w3c.dom.Element w_TaskNode = xml_Doc.createElement("task");
                    xml_RootNode.appendChild(w_TaskNode);

                    // Finish: computation XML attributes.
                    Traces.traceComputation("Project result, "
                        + "date = " + w_DateIt + ", "
                        + "reference = " + new Double(w_Computation.getReference()) + ", "
                        + "reference-risks = " + new Double(w_Computation.getReferenceRisks()) + ", "
                        + "reference-opportunities = " + new Double(w_Computation.getReferenceOpportunities()) + ", "
                        + "imputed = " + new Double(w_Computation.getImputed()) + ", "
                        + "remaining = " + new Double(w_Computation.getRemaining()) + ", "
                        + "remaining-risks = " + new Double(w_Computation.getRemainingRisks()) + ", "
                        + "remaining-opportunities = " + new Double(w_Computation.getRemainingOpportunities()) + ", "
                        + "planned = " + new Double(w_Computation.getPlanned()) + ", "
                        + "earned = " + new Double(w_Computation.getEarned())
                    );
                    //w_TaskNode.setAttribute("id", m_wMainTask.getId());   // @project attribute already set.
                    w_TaskNode.setAttribute("date", w_DateIt.toString());
                    w_TaskNode.setAttribute("reference", new Double(w_Computation.getReference()).toString());
                    w_TaskNode.setAttribute("reference-risks", new Double(w_Computation.getReferenceRisks()).toString());
                    w_TaskNode.setAttribute("reference-opportunities", new Double(w_Computation.getReferenceOpportunities()).toString());
                    w_TaskNode.setAttribute("imputed", new Double(w_Computation.getImputed()).toString());
                    w_TaskNode.setAttribute("remaining", new Double(w_Computation.getRemaining()).toString());
                    w_TaskNode.setAttribute("remaining-risks", new Double(w_Computation.getRemainingRisks()).toString());
                    w_TaskNode.setAttribute("remaining-opportunities", new Double(w_Computation.getRemainingOpportunities()).toString());
                    w_TaskNode.setAttribute("planned", new Double(w_Computation.getPlanned()).toString());
                    w_TaskNode.setAttribute("earned", new Double(w_Computation.getEarned()).toString());
                }
            }

            // Finish.
            return xml_Doc;
        } catch (Exception e) {
            Display.printStackTrace(e);
            return null;
        }
    }

    /** CLI error handler. Nothing special to do here, the errors should be notified by the CLI itself.
        @param location CLI error location information.
        @param message CLI error message. */
    public void onError(cli.ResourceString location, cli.ResourceString message) {
    }

    /** CLI exit handler. */
    public void onExit() {
    }

    private String m_jProjectName;
    private final Task m_wMainTask;
    private final Task m_wOutOfScopeImputations;
    private Task m_wCurrentTask;

    private User m_wCurrentUser;
    private Date m_wCurrentRevisionDate;
    private User.Map m_wUserMap;

}
